import numpy as np
from matplotlib.patches import Wedge, Polygon, RegularPolygon
from matplotlib.collections import PatchCollection
import matplotlib.pyplot as plt


class FootprintGeometry():

    def __init__(self):
        self._patches = []
        self._xlim = 0
        self._zenithAngleInDeg = 0
        self._lgs = []
        self._ngs = []
        self._targets = []
        self._dms = []
        self._instrFoVInArcsec = None
        self._layerAltitudeInMeter = 10000
        self._telescopeRadiusInMeter = 4.1
        self._drawMetapupil = True

    def setInstrumentFoV(self, instrFoVInArcsec, instrRotInDeg=0):
        self._instrFoVInArcsec = instrFoVInArcsec
        self._instrRotInDeg = instrRotInDeg

    def set_zenith_angle(self, zenithAngleInDeg):
        self._zenithAngleInDeg = zenithAngleInDeg

    def setLayerAltitude(self, altitudeInMeter):
        self._layerAltitudeInMeter = altitudeInMeter

    def setTelescopeRadiusInMeter(self, radiusInMeter):
        self._telescopeRadiusInMeter = radiusInMeter

    def addLgs(self,
               thetaSkyInArcsec,
               AzSkyInDeg):
        self._lgs.append(
            {'lRo': 0,
             'lAz': 0,
             'skyTheta': thetaSkyInArcsec,
             'skyAz': AzSkyInDeg})

    def addNgs(self, thetaSkyInArcsec, AzSkyInDeg):
        self._ngs.append(
            {'lRo': 0,
             'lAz': 0,
             'skyTheta': thetaSkyInArcsec,
             'skyAz': AzSkyInDeg})

    def addTarget(self, thetaSkyInArcsec, AzSkyInDeg):
        self._targets.append(
            {'lRo': 0,
             'lAz': 0,
             'skyTheta': thetaSkyInArcsec,
             'skyAz': AzSkyInDeg})

    def addDm(self, pitchOnLayer, nActs, size, rotationAngleInDeg):
        self._dms.append(
            {'lRo': 0,
             'lAz': 0,
             'pitch': pitchOnLayer,
             'nActs': nActs,
             'size': size,
             'skyRot': rotationAngleInDeg})

    def compute(self):
        self._lgsL = []
        if self._lgs:
            for l in self._lgs:
                ll = self._polToRect(l['lRo'], l['lAz'])
                cc = self._centerOffset(l['skyTheta'], l['skyAz'],
                                        self._layerAltitudeInMeter)
                ra = self._lgsRadius()
                self._lgsL.append(FootprintXYRadius(
                    ll[0] + cc[0], ll[1] + cc[1], ra,
                    self._layerAltitudeInMeter))

        self._ngsL = []
        if self._ngs:
            for l in self._ngs:
                cc = self._centerOffset(l['skyTheta'], l['skyAz'],
                                        self._layerAltitudeInMeter)
                ra = self._telescopeRadiusInMeter
                self._ngsL.append(FootprintXYRadius(cc[0], cc[1], ra,
                                                    self._layerAltitudeInMeter))

        self._targetsL = []
        for l in self._targets:
            cc = self._centerOffset(l['skyTheta'], l['skyAz'],
                                    self._layerAltitudeInMeter)
            ra = self._telescopeRadiusInMeter
            self._targetsL.append(FootprintXYRadius(cc[0], cc[1], ra,
                                                    self._layerAltitudeInMeter))

        self._sciL = []
        if self._instrFoVInArcsec:
            for sciFovCornerDeg in np.array([45, 135, 225, 315]
                                            ) + self._instrRotInDeg:
                cc = self._centerOffset(
                    self._instrFoVInArcsec / 2 * np.sqrt(2),
                    sciFovCornerDeg,
                    self._layerAltitudeInMeter)
                ra = self._telescopeRadiusInMeter
                self._sciL.append(
                    np.array([cc[0], cc[1], ra, sciFovCornerDeg]))

        self._metapupilL = []
        if self._lgs:
            raLgs = np.max([np.linalg.norm((l.x, l.y)) + l.r
                            for l in self._lgsL])
        else:
            raLgs = 0
        if self._ngs:
            raNgs = np.max([np.linalg.norm((l.x, l.y)) + l.r
                            for l in self._ngsL])
        else:
            raNgs = 0
        if self._targetsL:
            raTargets = np.max([np.linalg.norm((l.x, l.y)) + l.r
                                for l in self._targetsL])
        else:
            raTargets = 0
        ra = np.max([raLgs, raNgs, raTargets])

        self._metapupilL.append(FootprintXYRadius(0, 0, ra,
                                                  self._layerAltitudeInMeter))

        self._dmsL = []
        if self._dms:
            for l in self._dms:
                self._dmsL.append(l)
#                 ll = self._polToRect(l['lRo'], l['lAz'])
#                 cc = self._centerOffset(l['skyTheta'], l['skyAz'],
#                                         self._layerAltitudeInMeter)
#                 ra = self._lgsRadius()
#                 self._dmsL.append(FootprintXYRadius(
#                     ll[0] + cc[0], ll[1] + cc[1], ra))

    def getNgsFootprint(self):
        return self._ngsL

    def getLgsFootprint(self):
        return self._lgsL

    def getTargetFootprint(self):
        return self._targetsL

    def getMetapupilFootprint(self):
        return self._metapupilL

    # def getLayerAltitudeInMeter(self):
    #     return self._layerAltitudeInMeter

    def _computePatches(self):
        self._patches = []
        self._xlim = 0
        for l in self._lgsL:
            self._addAnnularFootprint(
                l.x, l.y, l.r, 0.99 * l.r, color='y', alpha=0.5)
            self._addAnnularFootprint(
                l.x, l.y, l.r, 0.00 * l.r, color='y', alpha=0.1)
        for l in self._ngsL:
            self._addAnnularFootprint(
                l.x, l.y, l.r, 0.99 * l.r, color='r', alpha=0.5)
        for l in self._targetsL:
            self._addAnnularFootprint(
                l.x, l.y, l.r, 0.99 * l.r, color='b', alpha=0.5)
        for l in self._sciL:
            self._addAnnularFootprint(
                l[0], l[1], l[2], 0.99 * l[2],
                theta1=l[3] - 10, theta2=l[3] + 10, color='b')
        if self._drawMetapupil:
            for l in self._metapupilL:
                self._addAnnularFootprint(
                    l.x, l.y, l.r, 0.99 * l.r, color='k')
        for l in self._dmsL:
            self._addDmFootprint(l, color='k', alpha=0.3)

    def scienceFieldRadius(self, rTel, fovInArcsec, hInMeter):
        return rTel + fovInArcsec / 2 * 4.848e-6 * hInMeter

    def _polToRect(self, ro, azInDeg):
        azInRad = azInDeg * np.pi / 180
        return ro * np.array(
            [np.cos(azInRad), np.sin(azInRad)])

    def _centerOffset(self, thetaInArcsec, azInDeg, hInMeter):
        return self._polToRect(
            thetaInArcsec * 4.848e-6 * hInMeter, azInDeg)

    def _lgsDistance(self):
        return 90000 / np.cos(self._zenithAngleInDeg * np.pi / 180)

    def _lgsRadius(self):
        return self._telescopeRadiusInMeter * (
            1 - self._layerAltitudeInMeter / self._lgsDistance())

    def _addAnnularFootprint(
            self, centerX, centerY,
            radiusOut, radiusIn=0,
            theta1=0, theta2=360,
            color='b', alpha=1):
        center = np.array([centerX, centerY])
        self._patches.append(
            Wedge(center, radiusOut, theta1, theta2,
                  width=(radiusOut - radiusIn), color=color, alpha=alpha))
        self._xlim = np.maximum(self._xlim,
                                np.max(np.abs(center) + radiusOut))

    def _addDmFootprint(self, l, color, alpha):
        rotAngleRad = l['skyRot'] * np.pi / 180
        for i in np.arange(-l['nActs'], l['nActs'] + 1):
            x = i * l['pitch']
            for j in np.arange(-l['nActs'], l['nActs'] + 1):
                y = j * l['pitch']
                xR = x * np.cos(rotAngleRad) - y * np.sin(rotAngleRad)
                yR = x * np.sin(rotAngleRad) + y * np.cos(rotAngleRad)
                self._addCrossFootprint(
                    xR, yR, l['size'], rotAngleRad, color, alpha)

    def _addCrossFootprint(
            self, centerX, centerY, size,
            thetaAngle, color, alpha):
        numVertices = 5
        center = np.array([centerX, centerY])
        self._patches.append(
            RegularPolygon(center, numVertices, size, thetaAngle,
                           color=color, alpha=alpha))
        # self._xlim = np.maximum(self._xlim,
        #                        np.max(np.abs(center) + size))

    def _addRectangle(
            self, centerX, centerY,
            sideX, sideY,
            color='b', alpha=0.1):
        bl = np.array([int(centerX - sideX // 2), int(centerY - sideY // 2)])
        tl = np.array([int(centerX - sideX // 2), int(centerY + sideY // 2)])
        br = np.array([int(centerX + sideX // 2), int(centerY - sideY // 2)])
        tr = np.array([int(centerX + sideX // 2), int(centerY + sideY // 2)])
        vertexes = np.vstack((bl, tl, tr, br))
        self._patches.append(
            Polygon(vertexes, closed=True, color=color, alpha=alpha))
        self._xlim = np.maximum(
            self._xlim,
            np.linalg.norm(np.array([centerX, centerY])) + 
            0.5 * np.linalg.norm(np.array([sideX, sideY])))

    def plot(self):
        self._computePatches()
        fig, ax = plt.subplots()
        ax.set_aspect('equal')
        ax.set_xlim(-self._xlim, self._xlim)
        ax.set_ylim(-self._xlim, self._xlim)
        p = PatchCollection(self._patches, match_original=True)
        ax.add_collection(p)
        ax.set_title('Footprints @ %g km, zenith= %7.3f°' % (
            self._layerAltitudeInMeter / 1000, self._zenithAngleInDeg))
        if self._instrFoVInArcsec:
            plt.text(0.5, 0.05,
                     'Science FoV %dx%d"' % (
                         self._instrFoVInArcsec,
                         self._instrFoVInArcsec),
                     color='b',
                     horizontalalignment='center',
                     verticalalignment='center',
                     transform=ax.transAxes)
        if self._lgs:
            plt.text(0.15, 0.05,
                     'LGS @ %d"' % self._lgs[0]['skyTheta'],
                     color='y',
                     horizontalalignment='center',
                     verticalalignment='center', transform=ax.transAxes)
        if self._ngs:
            plt.text(0.85, 0.05,
                     'NGS @ %d"' % self._ngs[0]['skyTheta'],
                     color='r',
                     horizontalalignment='center',
                     verticalalignment='center', transform=ax.transAxes)
        plt.show()

    def report(self):
        if self._lgs:
            for l in self._lgsL:
                print("LGS (x,y,r): %f, %f - %f" % (l.x, l.y, l.r))
        if self._ngs:
            for l in self._ngsL:
                print("NGS (x,y,r): %f, %f - %f" % (l.x, l.y, l.r))
        if self._targets:
            for l in self._targetsL:
                print("Targets (x,y,r): %f, %f - %f" % (l.x, l.y, l.r))
        for l in self._sciL:
            print("Science (x,y,r): %f, %f - %f" % (l[0], l[1], l[2]))
        for l in self._metapupilL:
            print("Metapupil (x,y,r): %f, %f - %f" % (l.x, l.y, l.r))


def mainFootprintGeometry(h=12000, lgsTh=15, lgsN=4, ngsTh=60, ngsN=3,
                          sciFov=20,
                          targets=[[0, 0], [60 * 1.414, 45]],
                          rTel=4.1, zenith_angle=0,
                          dm_pitch=1.0, dm_acts=10,
                          dm_act_size=0.1, dm_rot_angle=0):
    fg = FootprintGeometry()
    fg.setTelescopeRadiusInMeter(rTel)
    fg.set_zenith_angle(zenith_angle)
    fg.setInstrumentFoV(sciFov)
    fg.setLayerAltitude(h)
    if lgsN > 0:
        for azAng in np.arange(0, 360, 360 / lgsN):
            fg.addLgs(lgsTh, azAng)
    if ngsN > 0:
        for azAng in np.arange(0, 360, 360 / ngsN):
            fg.addNgs(ngsTh, azAng)
    for t in targets:
        fg.addTarget(t[0], t[1])
    fg.addDm(dm_pitch, dm_acts, dm_act_size, dm_rot_angle)
    fg.compute()
    fg.plot()
    fg.report()
    return fg


class FootprintXYRadius():
    '''
    Return (x, y) coordinates and radius of a footprint at altitude z.
    The altitude is also returned. 
    '''

    def __init__(self, x, y, r, z):
        self.x = x
        self.y = y
        self.r = r
        self.z = z




import sys
from io import StringIO
from contextlib import contextmanager


@contextmanager
def capture_output():
    '''
    Context manager to capture stdout and stderr.

    Captures stddout and stderr from a Python code section,
    using two StringIO objects.
    Example::

      with capture_output() as (out, err):
          routine_that_prints_lots()

    *out.getvalue()* will return as string with whatever
    was printed on stdout. *err.getvalue()* will return
    the same for stderr.
    Nothing will appear on screen.

    '''
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err



import numpy as np
from arte.utils.discrete_fourier_transform import \
    BidimensionalFourierTransform as bfft
from arte.types.scalar_bidimensional_function import \
    ScalarBidimensionalFunction as S2DF
from arte.types.domainxy import DomainXY


def logShow(image):
    from matplotlib import colors, cm, pyplot as plt

    norm = colors.LogNorm(vmin=image.max() / 1e6,
                          vmax=image.max(), clip='True')
    plt.imshow(image + 1e-16, cmap=cm.gray, norm=norm, origin="lower")


class FourierAdaptiveOptics(object):

    RAD2ARCSEC = (3600 * 180) / np.pi
    ARCSEC2RAD = np.pi / (180 * 3600)

    def __init__(self,
                 pupilDiameterInMeters=8.0,
                 wavelength=1e-6,
                 focalPlaneFieldOfViewInArcsec=1.0,
                 resolutionFactor=2):
        self._pupilDiameterInMeters = pupilDiameterInMeters
        self._wavelength = wavelength
        self._focalPlaneFieldOfViewInArcsec = focalPlaneFieldOfViewInArcsec
        self._resolutionFactor = resolutionFactor

        self._nyquistFocalPlanePixelSizeInArcsec = 0.5 * \
            self._wavelength / self._pupilDiameterInMeters * self.RAD2ARCSEC

        self._focalPlanePixelSizeInArcsec = \
            self._nyquistFocalPlanePixelSizeInArcsec / self._resolutionFactor

        self._computePixelSizes(self._focalPlanePixelSizeInArcsec,
                                self._focalPlaneFieldOfViewInArcsec,
                                self._wavelength)

        self._pupilDiameterInPixels = self._noFocalPlanePixels
        self._phaseMapInMeters = None

        self._focalPlaneCoordinatesInArcsec = \
            self._createFocalPlaneCoordinatesInArcSec()
        self._focalPlaneAngularFreqCoords = \
            self._createFocalPlaneAngularFrequencyCoordinatesInInverseRadians()
        self._pupilPlaneCoordinatesInMeters = \
            self._createPupilPlaneCoordinatesInMeters()
        self._pupilPlaneSpatialFrequencyCoordinatesInInverseMeters = \
            self._createPupilPlaneSpatialFrequencyCoordiantesInInverseMeters()

        self._resetAll()

        self._mask = None
        self._focalLength = 125.
        self._createCircularMask()
        self._createPupilFunction()

        self.setPhaseMapInMeters(self._createFlatPhaseMap())

    def _computePixelSizes(self,
                           focalPlanePixelSizeInArcsec,
                           focalPlaneFieldOfViewInArcsec,
                           wavelenghtInMeters):
        self._noFocalPlanePixels = int(
            focalPlaneFieldOfViewInArcsec / focalPlanePixelSizeInArcsec)

        self._focalPlaneAngularFrequencyPixelSizeInInverseRadians = 1. / (
            self._noFocalPlanePixels * focalPlanePixelSizeInArcsec *
            self.ARCSEC2RAD)

        self._pupilPlaneSpatialFrequencyPizelSizeInInverseMeters = (
            focalPlanePixelSizeInArcsec * self.ARCSEC2RAD /
            wavelenghtInMeters)

        self._pupilPlanePixelSizeInMeters = 1. / (
            self._noFocalPlanePixels *
            self._pupilPlaneSpatialFrequencyPizelSizeInInverseMeters)

    def _resetAll(self):
        self._field = None
        self._pupilFunction = None
        self._amplitudeTransferFunction = None
        self._psf = None
        self._otf = None
        self._stf = None

    def pupilDiameterInMeters(self):
        return self._pupilDiameterInMeters

    def wavelengthInMeters(self):
        return self._wavelength

    def resolutionFactor(self):
        return self._resolutionFactor

    def focalPlaneFieldOfViewInArcsec(self):
        return self._focalPlaneFieldOfViewInArcsec

    def focalPlaneSizeInPixels(self):
        return self._noFocalPlanePixels

    def focalPlanePixelSizeInArcsec(self):
        return self._focalPlanePixelSizeInArcsec

    def focalPlaneAngularFrequencyPixelSizeInInverseRadians(self):
        return self._focalPlaneAngularFrequencyPixelSizeInInverseRadians

    def pupilPlanePixelSizeInMeters(self):
        return self._pupilPlanePixelSizeInMeters

    def focalPlaneAngularFrequencyCoordinatesInInverseRadians(self):
        return self._focalPlaneAngularFreqCoords

    def focalPlaneCoordinatesInArcsec(self):
        return self._focalPlaneCoordinatesInArcsec

    def pupilPlaneCoordinatesInMeters(self):
        return self._pupilPlaneCoordinatesInMeters

    def pupilPlaneSpatialFrequencyPizelSizeInInverseMeters(self):
        return self._pupilPlaneSpatialFrequencyPizelSizeInInverseMeters

    def field(self):
        if self._field is None:
            self._computeField()
        return self._field

    def psf(self):
        if self._psf is None:
            self._createPsf()
        return self._psf

    def otf(self):
        if self._otf is None:
            self._createOtf()
        return self._otf

    def stf(self):
        if self._stf is None:
            self._createStructureFunction()
        return self._stf

    def pupilFunction(self):
        if self._pupilFunction is None:
            self._createPupilFunction()
        return self._pupilFunction

    def amplitudeTransferFunction(self):
        if self._amplitudeTransferFunction is None:
            self._createAmplitudeTransferFunction()
        return self._amplitudeTransferFunction

    def focalPlaneCoordsInArcsec(self):
        return self._focalPlaneCoordinatesInArcsec / 4.848e-6

    def setPhaseMapInMeters(self, phaseMapInMeters):
        self._phaseMapInMeters = phaseMapInMeters
        self._resetAll()

    def _createCircularMask(self):
        from arte.types.mask import CircularMask
        radiusInPx = 0.5 * self._pupilDiameterInMeters / \
            self.pupilPlanePixelSizeInMeters()
        centerInPx = [self._pupilDiameterInPixels / 2,
                      self._pupilDiameterInPixels / 2]
        self._mask = CircularMask(
            (self._pupilDiameterInPixels, self._pupilDiameterInPixels),
            radiusInPx, centerInPx)

    def _createPupilFunction(self):
        domain = DomainXY.from_shape((self._pupilDiameterInPixels,
                                      self._pupilDiameterInPixels),
                                     self.pupilPlanePixelSizeInMeters())
        self._pupilFunction = S2DF(
            self._mask.asTransmissionValue(), domain=domain)

    def _createFlatPhaseMap(self):
        nPx = self._pupilDiameterInPixels
        pupPxSize = self.pupilPlanePixelSizeInMeters()
        domain = DomainXY.from_shape((nPx, nPx), pupPxSize)
        phase = S2DF(np.ones((nPx, nPx)), domain=domain)
        return phase

    def _computeField(self):
        phaseInRadians = self._phaseMapInMeters.values / \
            self._wavelength * 2 * np.pi
        amplitude = np.ones_like(phaseInRadians)
        field = amplitude * np.exp(phaseInRadians * 1j)
        self._field = S2DF(
            field * self._mask.asTransmissionValue(),
            self._phaseMapInMeters.xmap,
            self._phaseMapInMeters.ymap)

    def _extendFieldMap(self, fm, howManyTimes):
        fmExt = np.zeros(np.array(fm.shape) * howManyTimes,
                         dtype=np.complex128)
        fmExt[0:fm.shape[0], 0:fm.shape[1]] = fm
        return fmExt

    def _createAmplitudeTransferFunction(self):
        pupF = self.pupilFunction()
        rescaleCoordFact = 1 / (self.wavelengthInMeters() * self._focalLength)
        self._amplitudeTransferFunction = S2DF(
            pupF.values,
            pupF.xmap * rescaleCoordFact,
            pupF.ymap * rescaleCoordFact)

    def _createOtf(self):
        ac = self._autoCorrelate(self.amplitudeTransferFunction())
        self._otf = S2DF(ac.values / ac.values.max(),
                         ac.xmap,
                         ac.ymap)

    def _autoCorrelate(self, scalar2dfunct):
        functFT = bfft.direct(scalar2dfunct)
        aa = S2DF(np.abs(functFT.values ** 2),
                  functFT.xmap,
                  functFT.ymap)
        return bfft.reverse(aa)

    def _createPsf(self):
        psf = bfft.reverse(self.otf())
        rescaleCoordFact = 1 / self._focalLength
        self._psf = S2DF(np.abs(psf.values),
                         psf.xmap * rescaleCoordFact,
                         psf.ymap * rescaleCoordFact)

    def _createStructureFunction(self):
        extFieldMap = self._extendFieldMap(self.field(),
                                           self._resolutionFactor)
        ac = self._autoCorrelate(extFieldMap)
        cc = (np.asarray(ac.shape) / 2).astype(np.int64)
        self._stf = 2 * (ac[cc[0], cc[1]] - ac)

    def _createFocalPlaneCoordinatesInArcSec(self):
        return bfft.distances_x_map(
            self.focalPlaneSizeInPixels(),
            self.focalPlanePixelSizeInArcsec())

    def _createFocalPlaneAngularFrequencyCoordinatesInInverseRadians(self):
        return bfft.frequencies_x_map(
            self.focalPlaneSizeInPixels(),
            self.focalPlanePixelSizeInArcsec() * self.ARCSEC2RAD)

    def _createPupilPlaneCoordinatesInMeters(self):
        return bfft.distances_x_map(
            self.focalPlaneSizeInPixels(),
            self.pupilPlanePixelSizeInMeters())

    def _createPupilPlaneSpatialFrequencyCoordiantesInInverseMeters(self):
        return bfft.frequencies_x_map(
            self.focalPlaneSizeInPixels(),
            self.pupilPlanePixelSizeInMeters())


class TurbulentPhase(object):

    def __init__(self):
        pass

    def correlationFunction(self):
        pass

    def structureFunction(self):
        pass

    def vonKarmanPowerSpectralDensity(self, r0, L0, frequency):
        return 0.0228 * (1. / r0) ** (5. / 3) * (frequency ** 2 + 1. / L0) ** (-11. / 6)

    def kolmogorovStructureFunction(self, r0, ro):
        return 6.88 * (ro / r0) ** (5. / 3)

    def dist(self, npx):
        pass



"""

From: http://code.activestate.com/recipes/278731-creating-a-daemon-the-python-way/

Disk And Execution MONitor (Daemon)

Configurable daemon behaviors:

   1.) The current working directory set to the "/" directory.
   2.) The current file creation mode mask set to 0.
   3.) Close all open files (1024). 
   4.) Redirect standard I/O streams to "/dev/null".

A failed call to fork() now raises an exception.

References:
   1) Advanced Programming in the Unix Environment: W. Richard Stevens
   2) Unix Programming Frequently Asked Questions:
         http://www.erlenstar.demon.co.uk/unix/faq_toc.html

who                when        what
-----------------  ----------  ----------
Chad J. Schroeder  2008-10-01  Created
Alfio Puglisi      2015        Changed WORKDIR and REDIRECT_TO
                               to be parameters of createDeamon()
Alfio Puglisi      2019-04-15  Converted to Python3

"""

__author__ = "Chad J. Schroeder"
__copyright__ = "Copyright (C) 2005 Chad J. Schroeder"

__revision__ = "$Id: daemonize.py,v 1.3 2008/10/01 13:33:46 labot Exp $"
__version__ = "0.2"

# Standard Python modules.
import os               # Miscellaneous OS interfaces.
import sys              # System-specific parameters and functions.

# Default daemon parameters.
# File mode creation mask of the daemon.
UMASK = 0

# Default maximum for the number of available file descriptors.
MAXFD = 1024

def createDaemon( WORKDIR, REDIRECT_TO = "/dev/null"):

   """Detach a process from the controlling terminal and run it in the
   background as a daemon.
   """

   try:
      # Fork a child process so the parent can exit.  This returns control to
      # the command-line or shell.  It also guarantees that the child will not
      # be a process group leader, since the child receives a new process ID
      # and inherits the parent's process group ID.  This step is required
      # to insure that the next call to os.setsid is successful.
      pid = os.fork()
   except OSError as e:
      raise Exception("%s [%d]" % (e.strerror, e.errno))

   if (pid == 0):   # The first child.
      # To become the session leader of this new session and the process group
      # leader of the new process group, we call os.setsid().  The process is
      # also guaranteed not to have a controlling terminal.
      os.setsid()

      # Is ignoring SIGHUP necessary?
      #
      # It's often suggested that the SIGHUP signal should be ignored before
      # the second fork to avoid premature termination of the process.  The
      # reason is that when the first child terminates, all processes, e.g.
      # the second child, in the orphaned group will be sent a SIGHUP.
      #
      # "However, as part of the session management system, there are exactly
      # two cases where SIGHUP is sent on the death of a process:
      #
      #   1) When the process that dies is the session leader of a session that
      #      is attached to a terminal device, SIGHUP is sent to all processes
      #      in the foreground process group of that terminal device.
      #   2) When the death of a process causes a process group to become
      #      orphaned, and one or more processes in the orphaned group are
      #      stopped, then SIGHUP and SIGCONT are sent to all members of the
      #      orphaned group." [2]
      #
      # The first case can be ignored since the child is guaranteed not to have
      # a controlling terminal.  The second case isn't so easy to dismiss.
      # The process group is orphaned when the first child terminates and
      # POSIX.1 requires that every STOPPED process in an orphaned process
      # group be sent a SIGHUP signal followed by a SIGCONT signal.  Since the
      # second child is not STOPPED though, we can safely forego ignoring the
      # SIGHUP signal.  In any case, there are no ill-effects if it is ignored.
      #
      # import signal           # Set handlers for asynchronous events.
      # signal.signal(signal.SIGHUP, signal.SIG_IGN)

      try:
         # Fork a second child and exit immediately to prevent zombies.  This
         # causes the second child process to be orphaned, making the init
         # process responsible for its cleanup.  And, since the first child is
         # a session leader without a controlling terminal, it's possible for
         # it to acquire one by opening a terminal in the future (System V-
         # based systems).  This second fork guarantees that the child is no
         # longer a session leader, preventing the daemon from ever acquiring
         # a controlling terminal.
         pid = os.fork()    # Fork a second child.
      except OSError as e:
         raise Exception("%s [%d]" % (e.strerror, e.errno))

      if (pid == 0):    # The second child.
         # Since the current working directory may be a mounted filesystem, we
         # avoid the issue of not being able to unmount the filesystem at
         # shutdown time by changing it to the root directory.
         os.chdir(WORKDIR)
         # We probably don't want the file mode creation mask inherited from
         # the parent, so we give the child complete control over permissions.
         os.umask(UMASK)
      else:
         # exit() or _exit()?  See below.
         os._exit(0)    # Exit parent (the first child) of the second child.
   else:
      # exit() or _exit()?
      # _exit is like exit(), but it doesn't call any functions registered
      # with atexit (and on_exit) or any registered signal handlers.  It also
      # closes any open file descriptors.  Using exit() may cause all stdio
      # streams to be flushed twice and any temporary files may be unexpectedly
      # removed.  It's therefore recommended that child branches of a fork()
      # and the parent branch(es) of a daemon use _exit().
      os._exit(0)   # Exit parent of the first child.

   # Close all open file descriptors.  This prevents the child from keeping
   # open any file descriptors inherited from the parent.  There is a variety
   # of methods to accomplish this task.  Three are listed below.
   #
   # Try the system configuration variable, SC_OPEN_MAX, to obtain the maximum
   # number of open file descriptors to close.  If it doesn't exists, use
   # the default value (configurable).
   #
   # try:
   #    maxfd = os.sysconf("SC_OPEN_MAX")
   # except (AttributeError, ValueError):
   #    maxfd = MAXFD
   #
   # OR
   #
   # if (os.sysconf_names.has_key("SC_OPEN_MAX")):
   #    maxfd = os.sysconf("SC_OPEN_MAX")
   # else:
   #    maxfd = MAXFD
   #
   # OR
   #
   # Use the getrlimit method to retrieve the maximum file descriptor number
   # that can be opened by this process.  If there is not limit on the
   # resource, use the default value.
   #
   import resource# Resource usage information.
   maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
   if (maxfd == resource.RLIM_INFINITY):
      maxfd = MAXFD
  
   # Iterate through and close all file descriptors.
   for fd in range(0, maxfd):
      try:
         os.close(fd)
      except OSError:   # ERROR, fd wasn't open to begin with (ignored)
         pass

   # Redirect the standard I/O file descriptors to the specified file.  Since
   # the daemon has no controlling terminal, most daemons redirect stdin,
   # stdout, and stderr to /dev/null.  This is done to prevent side-effects
   # from reads and writes to the standard I/O file descriptors.

   # This call to open is guaranteed to return the lowest file descriptor,
   # which will be 0 (stdin), since it was closed above.
   os.open(REDIRECT_TO, os.O_RDWR | os.O_CREAT)  # standard input (0)

   # Duplicate standard input to standard output and standard error.
   os.dup2(0, 1)            # standard output (1)
   os.dup2(0, 2)            # standard error (2)

   return(0)

if __name__ == "__main__":

   retCode = createDaemon()

   # The code, as is, will create a new file in the root directory, when
   # executed with superuser privileges.  The file will contain the following
   # daemon related process parameters: return code, process ID, parent
   # process group ID, session ID, user ID, effective user ID, real group ID,
   # and the effective group ID.  Notice the relationship between the daemon's 
   # process ID, process group ID, and its parent's process ID.

   procParams = """
   return code = %s
   process ID = %s
   parent process ID = %s
   process group ID = %s
   session ID = %s
   user ID = %s
   effective user ID = %s
   real group ID = %s
   effective group ID = %s
   """ % (retCode, os.getpid(), os.getppid(), os.getpgrp(), os.getsid(0),
   os.getuid(), os.geteuid(), os.getgid(), os.getegid())

   open("createDaemon.log", "w").write(procParams + "\n")

   sys.exit(retCode)




import numpy as np
from arte.types.zernike_coefficients import ZernikeCoefficients


class ModalCoefficients(ZernikeCoefficients):
    FIRST_MODE = 0

    def __init__(self, coefficients, counter=0, first_mode=0):

        self._coefficients = coefficients
        self._counter = counter
        self.FIRST_MODE = first_mode

    def modeIndexes(self):
        return np.arange(self.FIRST_MODE, self.FIRST_MODE + self.numberOfModes())

    def numberOfModes(self):
        return len(self._coefficients)

    def getM(self, modeIndexes):
        return self.toNumpyArray()[np.array(modeIndexes) - self.FIRST_MODE]

    def toDictionary(self):
        keys = self.modeIndexes()
        values = self._coefficients
        return dict(list(zip(keys, values)))

    def toNumpyArray(self):
        return self._coefficients

    @staticmethod
    def fromNumpyArray(coefficientsAsNumpyArray, counter=0, **kwargs):
        return ModalCoefficients(np.array(coefficientsAsNumpyArray), counter, **kwargs)

    def counter(self):
        return self._counter

    def setCounter(self, counter):
        self._counter = counter

    def __eq__(self, o):
        if self._counter != o._counter:
            return False
        if not np.array_equal(self._coefficients, o._coefficients):
            return False
        return True

    def __ne__(self, o):
        return not self.__eq__(o)

    def __str__(self):
        return str(self._coefficients)





import astropy.units as u
from arte.time_series.indexer import ModeIndexer
from arte.dataelab.base_projection import BaseProjection


class BaseResidualModes(BaseProjection):
    '''
    Base class for residual modes

    slopes_timeseries: time series object with slopes data
    modalrec: data object with modal reconstructor
    '''

    def __init__(self, slopes_timeseries, modalrec, astropy_unit=u.m, data_label='Modal coefficients'):
        super().__init__(slopes_timeseries,
                         modalrec,
                         astropy_unit=astropy_unit,
                         data_label=data_label)
        self._nmodes = None   # Lazy initialization
        self._indexer = None

    def nmodes(self):
        '''Number of modes'''
        if self._nmodes is None:
            self._nmodes = self._projection_matrix.get_data().shape[0]
        return self._nmodes

    def get_index_of(self, *args, **kwargs):
        if self._indexer is None:
            self._indexer = ModeIndexer(max_mode=self.nmodes())
        return self._indexer.modes(*args, **kwargs)

    def slopes(self):
        '''Slopes timeseries'''
        return self._source_timeseries

    def modalrec(self):
        '''Modal reconstructor'''
        return self._projection_matrix

# __oOo__


import numpy as np


class Slopes(object):
    ''''
    Class that represents a generic WFS slope frame
    '''
    def __init__(self, slopesx_compressed, slopesy_compressed, mask2d):
        '''
        Parameters
        ----------
        slopesx_compressed: np.ndarray
            1D slope vector
        slopesy_compressed: np.ndarray
            1D slope vector
        mask: np.ndarray
            2D boolean mask to rearrange a 1D slope vector into a 2D frame,
            valid for both X and Y slope vectors.
        '''
        assert len(slopesx_compressed) == len(slopesy_compressed), \
            'X and Y slopes length differ'
        self._mask2d = mask2d
        self._vector = np.hstack((slopesx_compressed, slopesy_compressed))

    @staticmethod
    def from_2dmaps(mapX, mapY):
        '''
        Build a Slopes object from X/Y 2D maps or a series of X/Y 2D maps.
        Both maps must be np.ma.MaskedArray objects. Masks for X and Y slopes
        must be identical.

        Parameters
        ----------
        mapX: np.ma.MaskedArray
            2D slope map
        mapY: np.ma.MaskedArray
            2D slope map

        Returns
        -------
        slopes: Slopes object
            A new Slopes object. Valid slopes (as indicated by the mask)
            are copied into an internal array. No references to the input data are kept.
        '''
        assert isinstance(mapX, np.ma.MaskedArray), 'mapX is not a np.ma.MaskedArray'
        assert isinstance(mapY, np.ma.MaskedArray), 'mapY is not a np.ma.MaskedArray'
        assert mapX.shape == mapY.shape, 'mapX and mapY shapes differ'
        assert np.all(mapX.mask == mapY.mask), 'mapX and mapY masks differ'
        return Slopes(mapX.compressed(), mapY.compressed(), np.ma.getmaskarray(mapX))

    @staticmethod
    def fromNumpyArray(mapXAsMaskedNumpyArray,
                       mapYAsMaskedNumpyArray):
        return Slopes.from_2dmaps(mapXAsMaskedNumpyArray,
                                  mapYAsMaskedNumpyArray)

    def toNumpyArray(self):
        return self.mapX(), self.mapY()

    def numberOfSlopes(self):
        return len(self._vector)

    @property
    def shape2d(self):
        return self._mask2d.shape

    def _map2d(self, data1d):
        data2d = np.zeros_like(self._vector, shape=self._mask2d.shape)
        data2d[np.where(~self._mask2d)] = data1d
        return np.ma.array(data2d, mask=self._mask2d)

    def mapX(self):
        return self._map2d(self.vectorX())

    def mapY(self):
        return self._map2d(self.vectorY())

    def vector(self):
        return self._vector

    def vectorX(self):
        return self._vector[:self.numberOfSlopes() // 2]

    def vectorY(self):
        return self._vector[self.numberOfSlopes() // 2:]

    def __eq__(self, o):
        if not isinstance(o, Slopes):
            return False
        return np.array_equal(self._vector, o._vector)

    def __ne__(self, o):
        return not self.__eq__(o)




import abc
import os
import inspect


class LoggerException(Exception):
    pass


class AbstractLogger(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def fatal(self, m):
        raise RuntimeError("Do not use me!")

    @abc.abstractmethod
    def error(self, m):
        raise RuntimeError("Do not use me!")

    @abc.abstractmethod
    def warn(self, m):
        raise RuntimeError("Do not use me!")

    @abc.abstractmethod
    def notice(self, m):
        raise RuntimeError("Do not use me!")

    @abc.abstractmethod
    def debug(self, m):
        raise RuntimeError("Do not use me!")


class PythonLogger(AbstractLogger):

    def __init__(self, logger):
        self._logger= logger
        self._radix= ''


    def setRelativePathRadix(self, radix):
        self._radix= radix


    def _getRelativeDirName(self, fullPath):
        return os.path.relpath(fullPath, self._radix)


    def _getCallerTrace(self):
        stackCaller= inspect.stack()[3]
        return '%s:%s' % (self._getRelativeDirName(stackCaller[1]),
                          stackCaller[2])

    def _formatMessage(self, m):
        return '%s %s: %s' % (self._logger.name, self._getCallerTrace(), m)

    def fatal(self, m):
        msg= self._formatMessage(m)
        self._logger.fatal(msg)

    def error(self, m):
        msg= self._formatMessage(m)
        self._logger.error(msg)

    def warn(self, m):
        msg= self._formatMessage(m)
        self._logger.warning(msg)

    def notice(self, m):
        msg= self._formatMessage(m)
        self._logger.info(msg)

    def debug(self, m):
        msg= self._formatMessage(m)
        self._logger.debug(msg)


class ObservableLogger(AbstractLogger):


    def __init__(self, wrappedLogger):
        self._wrappedLogger= wrappedLogger
        self._listeners= []

    def fatal(self, m):
        for each in self._listeners:
            each.onFatal(m)
        self._wrappedLogger.fatal(m)


    def error(self, m):
        for each in self._listeners:
            each.onError(m)
        self._wrappedLogger.error(m)


    def warn(self, m):
        for each in self._listeners:
            each.onWarning(m)
        self._wrappedLogger.error(m)


    def notice(self, m):
        for each in self._listeners:
            each.onNotice(m)
        self._wrappedLogger.notice(m)


    def debug(self, m):
        for each in self._listeners:
            each.onDebug(m)
        self._wrappedLogger.debug(m)


    def addListener(self, listener):
        self._listeners.append(listener)


class LoggerListener(object):


    def onDebug(self, message):
        self._methodOfPureAbstractClass()


    def onNotice(self, message):
        self._methodOfPureAbstractClass()


    def onWarning(self, message):
        self._methodOfPureAbstractClass()


    def onError(self, message):
        self._methodOfPureAbstractClass()


    def onFatal(self, message):
        self._methodOfPureAbstractClass()


    def _methodOfPureAbstractClass(self):
        assert False, "Pure abstract class method"


class DummyLogger(AbstractLogger):

    def fatal(self, m):
        pass

    def error(self, m):
        pass

    def warn(self, m):
        pass

    def notice(self, m):
        pass

    def debug(self, m):
        pass


class AbstractLoggerFactory(object):

    def getLogger(self, loggerName):
        raise RuntimeError("Do not use me!")


class PythonLoggerFactory(AbstractLoggerFactory):

    def getLogger(self, name):
        import logging

        logger= logging.getLogger(name)
        pyLogger= PythonLogger(logger)
        radix= os.path.join(__file__, '..', '..')
        pyLogger.setRelativePathRadix(radix)
        return pyLogger


class DummyLoggerFactory(AbstractLoggerFactory):

    def getLogger(self, name):
        return DummyLogger()


class ObservableLoggerFactory(AbstractLoggerFactory):

    def __init__(self, wrappedLoggerFactory):
        self._wrappedLoggerFactory= wrappedLoggerFactory
        self._createdLoggerByName= {}
        self._listeners= []


    def _registerListenersTo(self, logger):
        for each in self._listeners:
            logger.addListener(each)


    def getLogger(self, name):
        if name in list(self._createdLoggerByName.keys()):
            return self._createdLoggerByName[name]

        wrappedLogger= self._wrappedLoggerFactory.getLogger(name)
        logger= ObservableLogger(wrappedLogger)
        self._createdLoggerByName[name]= logger

        self._registerListenersTo(logger)

        return logger


    def getLoggerMap(self):
        return self._createdLoggerByName


    def addListener(self, loggerListener):
        self._listeners.append(loggerListener)
        for loggerName in self._createdLoggerByName:
            logger= self._createdLoggerByName[loggerName]
            logger.addListener(loggerListener)


class Logger(object):

    DEFAULT_LOGGER_FACTORY= PythonLoggerFactory
    _loggerFactory= DEFAULT_LOGGER_FACTORY()


    @staticmethod
    def of(loggerName):
        return Logger._loggerFactory.getLogger(loggerName)


    @staticmethod
    def setLoggerFactory(loggerFactory):
        Logger._loggerFactory= loggerFactory



#
# (c) Mark Ivey 2008
#
# http://zovirl.com/2008/11/04/interpolated-lookup-tables-in-python/

class InterpolatedArray(object):

  """An array-like object that provides
  interpolated values between set points."""

  def __init__(self, points):
    self.points = sorted(points)

  def __getitem__(self, x):
    if x < self.points[0][0] or x > self.points[-1][0]:
      raise ValueError
    lower_point, upper_point = self._GetBoundingPoints(x)
    return self._Interpolate(x, lower_point, upper_point)

  def _GetBoundingPoints(self, x):
    """Get the lower/upper points that bound x."""
    lower_point = None
    upper_point = self.points[0]
    for point  in self.points[1:]:
      lower_point = upper_point
      upper_point = point
      if x <= upper_point[0]:
        break
    return lower_point, upper_point

  def _Interpolate(self, x, lower_point, upper_point):
    """Interpolate a Y value for x given lower & upper
    bounding points."""
    slope = (float(upper_point[1] - lower_point[1]) /
             (upper_point[0] - lower_point[0]))
    return lower_point[1] + (slope * (x - lower_point[0]))




import numpy as np
from scipy.special import binom



class ImageMoments():

    def __init__(self, image, shiftPixels=0.5):
        self._image= image
        self._shiftPixels= shiftPixels
        nrows, ncols = self._image.shape
        self.y, self.x = np.mgrid[:nrows, :ncols] + self._shiftPixels
        self._m00= self._image.sum()


    def rawMoment(self, iord, jord):
        data = self._image * self.x**iord * self.y**jord
        return data.sum()


    def centralNormalizedMoment(self, iord, jord):
        return self.central_moment(iord, jord) / self._m00


    # TODO: (lb) 20160104 make it faster if needed, it is easy!
    def central_moment(self, iord, jord):
        if iord == 0 and jord == 0:
            return self._m00
        if (iord == 0 and jord == 1) or (iord == 1 and jord == 0):
            return 0.
        moment=0.
        x_bar= self.xBar()
        y_bar= self.yBar()
        for m in np.arange(iord+1):
            for n in np.arange(jord+1):
                moment += binom(iord, m) * binom(jord, n) * \
                    (-x_bar)**(iord-m) * (-y_bar)**(jord-n) * \
                    self.rawMoment(m, n)
        return moment


    def xBar(self):
        return self.rawMoment(1, 0) / self.rawMoment(0, 0)


    def yBar(self):
        return self.rawMoment(0, 1) / self.rawMoment(0, 0)


    def centroid(self):
        return np.array([self.xBar(), self.yBar()])


    def covarianceMatrix(self):
        u11= self.centralNormalizedMoment(1, 1)
        u20= self.centralNormalizedMoment(2, 0)
        u02= self.centralNormalizedMoment(0, 2)
        cov = np.array([[u20, u11], [u11, u02]])
        return cov


    def eigenvalues(self):
        u11= self.centralNormalizedMoment(1, 1)
        u20= self.centralNormalizedMoment(2, 0)
        u02= self.centralNormalizedMoment(0, 2)
        aa= 0.5* (u20 + u02)
        bb= 0.5* np.sqrt(4* u11** 2 + (u20- u02)** 2)
        return np.array([aa+ bb, aa- bb])


    def eccentricity(self):
        eigen= self.eigenvalues()
        return np.sqrt(1- eigen[1]/ eigen[0])


    def orientation(self):
        u11= self.centralNormalizedMoment(1, 1)
        u20= self.centralNormalizedMoment(2, 0)
        u02= self.centralNormalizedMoment(0, 2)
        return 0.5 * np.arctan2(2* u11, u20- u02)


    def semiAxes(self):
        return 2* np.sqrt(self.eigenvalues())


    def principalMomentsOfInertia(self):
        return np.array([self.centralNormalizedMoment(2, 0),
                         self.centralNormalizedMoment(0, 2)])


    def effectiveRadius(self):
        return np.sqrt(np.product(self.semiAxes()))



# -*- coding: utf-8 -*-
#########################################################
#
# who       when        what
# --------  ----------  ---------------------------------
# apuglisi  2019-12-27  Created
#
#########################################################

from collections import OrderedDict


class TabularReport():
    '''
    Report with column-based format.

    Prints a new line of values each time it is triggered.

    Columns are defined at initialization with the *column_names*, and they
    can be set using the [] operator. Each column will occupy `column_width`
    characters (default 10 characters). Columns will be printed in the order
    they were added.

    Each column has a formatter function, which can set using the
    .fmt dictionary. The formatter function must take a single parameter,
    the value to be displayed, and return a string. By default, all columns
    are formatted with str().

    It is possible to add a column on the fly assigning
    to it using the [] operator. The new column will be assigned the
    default str() formatter.

    Every time the report() function is called, an internal counter is
    incremented and, if at least `decimation` counts have passed since
    the last time, the values are printed on stdout. By default, `decimation`
    is 1 and all lines are printed. A header with the column names is printed
    before the first line and after each `repeat_hdr` lines have been
    printed (default 100). If `add_iter` is True, the counter is printed
    as the first column.

    Attributes
    ----------
    decimation : int
        how many calls to report() are needed before a line is printed
    hdr_decimation: int
        how many lines are printed before the header is printed again
    add_iter: bool
        if True, the iteration number is added as the first column
    column_width: int
        number of characters assigned to each column
    iteration_hdr: str
        string used as column name for the iteration number
    ['column_name']
        Assign values to column using the [] operator.
    fmt['column_name']
        Assign special formatting functions using the fmt dictionary.

    Examples
    --------
    >>> from arte.utils.tabular_report import TabularReport
    >>> r = TabularReport(['pippo','pluto'])
    >>> for i in range(2):
    ...    r['pippo'] = i*2
    ...    r['pluto'] = i+10
    ...    r.report()
    >>> r['paperino'] = 'a'
    >>> r.fmt['paperino'] = ord
    >>> r.print_header()
    >>> r.report()

    Result::

     iteration  pippo      pluto
     ---------------------------------
     1          0          10
     2          2          11

     iteration  pippo      pluto      paperino
     --------------------------------------------
     3          2          11         97
    '''

    def __init__(self, column_names, decimation: int = 1,
                                     hdr_decimation: int = 100,
                                     add_iter: bool = True,
                                     column_width: int = 10,
                                     iteration_hdr='iteration'):

        self.values = OrderedDict.fromkeys(column_names)
        self.fmt = OrderedDict.fromkeys(column_names, str)

        self.decimation = decimation
        self.hdr_decimation = hdr_decimation
        self.add_iter = add_iter
        self.column_width = column_width
        self.iteration_hdr = iteration_hdr

        self.fmt[iteration_hdr] = str

        self.counter = 0
        self.hdr_counter = 0

    def __setitem__(self, key, value):
        if key not in self.fmt:
            self.fmt[key] = str
        self.values[key] = value

    def report(self):
        '''
        Increment the iteration counter and if, conditions match,
        print a header and/or a line of values.
        '''
        self.counter += 1
        if self.counter % self.decimation != 0:
            return

        if self.hdr_counter % self.hdr_decimation == 0:
            self.print_header()

        self.hdr_counter += 1
        self.print_line()

    def print_header(self):
        '''Unconditionally print a line with the header.'''

        cols = self.values.keys()
        if self.add_iter:
            cols = ['iteration'] + list(cols)

        fmt = '%%-%ds ' % self.column_width
        hdr = ''.join([fmt % col for col in cols])

        print()
        print(hdr)
        print('-' * len(hdr))

    def print_line(self):
        '''Unconditionally print a line with all the current values'''

        cols = self.values.keys()
        formatted_values = [self.fmt[col](self.values[col]) for col in cols]

        if self.add_iter:
            formatted_values = [str(self.counter)] + formatted_values

        fmt = '%%-%ds ' % self.column_width
        line = ''.join([fmt % value for value in formatted_values])

        print(line)

# ___oOo___



r''' This module manages atmospheric Cn2 profiles.

The class Cn2Profile is the base class: it allows to describe a profile
as a set of layers each one with specified J, outer scale :math:`L_0`,
wind speed and wind direction and located at a specified altitude h
above the ground

Cn2Profile allows to specify the zenith angle z and the wavelength and 

The following relations are used, where the index i identifies
the i-th layer and the airmass is :math:`X=\sec z`.


.. math::
   :nowrap:

   \begin{eqnarray}
    k & = & \frac{2 \pi}{\lambda} \\
    J_i & = & {C_n^2}_i \, dh_i  \\
    {r_0}_i & = & (0.423 \, X \, k^2 J_i)^{-3/5} \\
    r_0 & = & \big( \sum{{{r_0}_i}^{-5/3}} \big)^{-3/5} \\
        & = & \big( \sum{0.423 X k^2 J_i} \big)^{-3/5} \\
    \theta_0 & = & \big( 2.914 k^2 X^{8/3} \sum{J_i h_i^{5/3}} \big)^{-3/5} 
   \end{eqnarray}

'''
import numpy as np
from astropy.units.quantity import Quantity
from arte.utils.constants import DEG2RAD, RAD2ARCSEC
from arte.utils.package_data import dataRootDir
import os
from astropy.io import fits
import astropy.units as u
from astropy.utils.misc import isiterable


class Cn2Profile(object):
    """
    Parameters
    ----------
        layersJs: :class:`~numpy:numpy.ndarray`
                  array of layers Js in meters**(1/3)
        layersL0: :class:`~numpy:numpy.ndarray`
                  array of layers outer-scale L0 in meters
        layersAltitude: :class:`~numpy:numpy.ndarray`
                  array of layers altitude at Zenith in meters
        layersWindSpeed: :class:`~numpy:numpy.ndarray`
                  array of layers wind speed in meters per second
        layersWindDirection: :class:`~numpy:numpy.ndarray`
                  array of layers wind direction in degrees, clockwise
                  from North

    Every array must be 1D and have the same size, defining the number of
    layers of the profile

    All parameters must be defined at zenith
    """

    DEFAULT_LAMBDA = 500e-9
    DEFAULT_AIRMASS = 1.0

    def __init__(self,
                 layersJs,
                 layersL0,
                 layersAltitude,
                 layersWindSpeed,
                 layersWindDirection):
        self._layersJs = np.array(layersJs)
        self._layersL0 = np.array(layersL0)
        self._layersAltitudeInMeterAtZenith = np.array(layersAltitude)
        self._layersWindSpeed = np.array(layersWindSpeed)
        self._layersWindDirection = np.array(layersWindDirection)
        self._zenithInDeg = 0
        self._airmass = 1.0
        self._lambda = 500e-9
        self._checkSameShape(self._layersL0, self._layersJs, 'L0', 'Js')
        self._checkSameShape(self._layersAltitudeInMeterAtZenith,
                             self._layersJs, 'Altitude', 'Js')
        self._checkSameShape(self._layersWindSpeed, self._layersJs,
                             'WindSpeed', 'Js')
        self._checkSameShape(self._layersWindDirection, self._layersJs,
                             'WindDirection', 'Js')

    def _checkSameShape(self, a, b, nameA, nameB):
        assert a.shape == b.shape, '%s shape %s != %s shape %s' % (
            nameA, a.shape, nameB, b.shape)

    @classmethod
    def from_r0s(cls,
                 layersR0,
                 layersL0,
                 layersAltitude,
                 layersWindSpeed,
                 layersWindDirection):
        """
        Cn2 profile constructor from r0 values of each layer

        Parameters
        ----------
            layersR0: :class:`~numpy:numpy.ndarray`
                      array of layers r0 in meters at 500nm
            layersL0: :class:`~numpy:numpy.ndarray`
                      array of layers outer-scale L0 in meters
            layersAltitude: :class:`~numpy:numpy.ndarray`
                      array of layers altitude at Zenith in meters
            layersWindSpeed: :class:`~numpy:numpy.ndarray`
                      array of layers wind speed in meters per second
            layersWindDirection: :class:`~numpy:numpy.ndarray`
                      array of layers wind direction in degrees, clockwise
                      from North


        Every array must be 1D and have the same size, defining the number of
        layers of the profile

        All parameters must be defined at zenith
        """
        layersR0 = cls._quantitiesToValue(layersR0)
        layersL0 = cls._quantitiesToValue(layersL0)
        layersAltitude = cls._quantitiesToValue(layersAltitude)
        layersWindSpeed = cls._quantitiesToValue(layersWindSpeed)
        layersWindDirection = cls._quantitiesToValue(
            np.deg2rad(layersWindDirection))
        js = cls._r02js(layersR0, cls.DEFAULT_AIRMASS, cls.DEFAULT_LAMBDA)
        return Cn2Profile(js,
                          layersL0,
                          layersAltitude,
                          layersWindSpeed,
                          layersWindDirection)

    @classmethod
    def from_fractional_j(cls,
                          r0AtZenith,
                          layersFractionalJ,
                          layersL0,
                          layersAltitude,
                          layersWindSpeed,
                          layersWindDirection):
        """
        Cn2 profile constructor from total r0 at zenith and fractional J of
        each layer

        Parameters
        ----------
            r0AtZenith: float
                        overall r0 at zenith [m]
            layersFractionalJ: :class:`~numpy:numpy.ndarray`
                               array of J values for each layer.
                               Array must sum up to 1
            layersL0: :class:`~numpy:numpy.ndarray`
                      array of layers outer-scale L0 in meters
            layersAltitude: :class:`~numpy:numpy.ndarray`
                      array of layers altitude at Zenith in meters
            layersWindSpeed: :class:`~numpy:numpy.ndarray`
                      array of layers wind speed in meters per second
            layersWindDirection: :class:`~numpy:numpy.ndarray`
                      array of layers wind direction in degrees, clockwise
                      from North


        Every array must be 1D and have the same size, defining the number of
        layers of the profile

        All parameters must be defined at zenith

        """
        r0AtZenith = cls._quantitiesToValue(r0AtZenith)
        layersFractionalJ = cls._quantitiesToValue(layersFractionalJ)
        layersL0 = cls._quantitiesToValue(layersL0)
        layersAltitude = cls._quantitiesToValue(layersAltitude)
        layersWindSpeed = cls._quantitiesToValue(layersWindSpeed)
        layersWindDirection = cls._quantitiesToValue(
            np.deg2rad(layersWindDirection))
        sumJ = np.array(layersFractionalJ).sum()
        assert (np.abs(sumJ - 1.0) < 0.01), \
            "Total of J values must be 1.0, got %g" % sumJ
        totalJ = r0AtZenith ** (-5. / 3) / (
            0.422727 * (2 * np.pi / cls.DEFAULT_LAMBDA) ** 2 * 
            cls.DEFAULT_AIRMASS)
        js = np.array(layersFractionalJ) * totalJ
        return Cn2Profile(js,
                          layersL0,
                          layersAltitude,
                          layersWindSpeed,
                          layersWindDirection)

    @staticmethod
    def _quantitiesToValue(quantity):
        if isiterable(quantity):
            if isinstance(quantity[0], Quantity):
                value = np.array(
                    [quantity[i].si.value for i in range(len(quantity))])
            else:
                value = value = np.array(
                    [quantity[i] for i in range(len(quantity))])
        else:
            if isinstance(quantity, Quantity):
                value = quantity.si.value
            else:
                value = quantity
        return value

    @staticmethod
    def _r02js(r0s, airmass, wavelength):
        js = np.array(r0s) ** (-5. / 3) / (
            0.422727 * airmass * (2 * np.pi / wavelength) ** 2)
        return js

    @staticmethod
    def _js2r0(Js, airmass, wavelength):
        r0s = (0.422727 * airmass * (2 * np.pi / wavelength) ** 2 * 
               np.array(Js)) ** (-3. / 5)
        return r0s

    def set_zenith_angle(self, zenithAngleInDeg):
        if isinstance(zenithAngleInDeg, Quantity):
            value = zenithAngleInDeg.to(u.deg).value
        else:
            value = zenithAngleInDeg
        self._zenithInDeg = value
        self._airmass = self.zenith_angle_to_airmass(value)

    def zenith_angle(self):
        return self._zenithInDeg * u.deg

    @staticmethod
    def zenith_angle_to_airmass(zenithAngleInDeg):
        if isinstance(zenithAngleInDeg, Quantity):
            value = zenithAngleInDeg.to(u.deg).value
        else:
            value = zenithAngleInDeg
        zenithInRad = value * DEG2RAD
        return 1. / np.cos(zenithInRad)

    def airmass(self):
        '''
        Returns
        -------
        airmass: float
            airmass at specified zenith angle
        '''
        return self._airmass * u.dimensionless_unscaled

    def layers_distance(self):
        return self._layersAltitudeInMeterAtZenith * self._airmass * u.m

    def set_wavelength(self, wavelengthInMeters):
        if isinstance(wavelengthInMeters, Quantity):
            value = wavelengthInMeters.to(u.m).value
        else:
            value = wavelengthInMeters
        self._lambda = value

    def wavelength(self):
        return self._lambda * u.m

    def number_of_layers(self):
        return Quantity(len(self._layersJs), dtype=int)

    def fractional_j(self):
        return Quantity(self._layersJs)

    def set_wind_speed(self, windSpeed):
        if isinstance(windSpeed, Quantity):
            value = windSpeed.to(u.m / u.s).value
        else:
            value = windSpeed
        self._layersWindSpeed = value

    def wind_speed(self):
        """
        Returns:
            (array): windspeed of each layer in m/s
        """
        return self._layersWindSpeed * u.m / u.s

    def mean_wind_speed(self):
        return (
            np.sum(
                self._layersWindSpeed ** (5. / 3) * self._layersJs
            ) / np.sum(
                self._layersJs)
        ) ** (3. / 5) * u.m / u.s

    def set_wind_direction(self, windDirectionInDeg):
        if isinstance(windDirectionInDeg, Quantity):
            value = windDirectionInDeg.to(u.rad).value
        else:
            value = np.deg2rad(windDirectionInDeg)
        self._layersWindDirection = value

    def wind_direction(self):
        """
        Returns
        -------
        wind: :class:`~astropy:astropy.units.quantity.Quantity` containing an array.
            wind direction of each layer [deg, clockwise from N]
        """
        # TODO: check the unit. layersWindDirection is converted in SI, so
        # in radiants. I think that I must return np.rad2deg(layersWindDir) *
        # u.deg
        return np.rad2deg(self._layersWindDirection) * u.deg

    def seeing(self):
        """
        Returns
        -------
        seeing: :class:`~astropy:astropy.units.quantity.Quantity` equivalent to arcsec
            seeing value at specified lambda and zenith angle
            defined as 0.98 * lambda / r0
        """
        return 0.98 * self.wavelength() / self.r0() * RAD2ARCSEC * \
            u.arcsec

    def r0(self):
        '''
        Returns
        -------
        r0: :class:`~astropy:astropy.units.quantity.Quantity` equivalent to meters
            Fried parameter at defined wavelength and zenith angle
        '''
        return (0.422727 * self._airmass * 
                (2 * np.pi / self._lambda) ** 2 * 
                np.sum(self._layersJs)) ** (-3. / 5) * u.m

    def r0s(self):
        '''
        Returns
        -------
        r0: :class:`~astropy:astropy.units.quantity.Quantity` equivalent to meters of :class:`~numpy:numpy.ndarray`
            Fried parameter of each layer at defined wavelength and zenith angle 
        '''
        return self._js2r0(self._layersJs,
                           self._airmass,
                           self._lambda) * u.m

    def theta0(self):
        '''
        Returns:
            theta0 (float): isoplanatic angle at specified lambda and zenith
                angle [arcsec]
        '''
        return (2.914 * self._airmass ** (8. / 3) * 
                (2 * np.pi / self._lambda) ** 2 * 
                np.sum(self._layersJs * 
                       self._layersAltitudeInMeterAtZenith ** (5. / 3))
                ) ** (-3. / 5) * RAD2ARCSEC * u.arcsec

    def tau0(self):
        '''
        Returns:
            tau0 (float): tau0 at specified lambda and zenith angle [sec]
        '''
        return (2.914 * self._airmass * 
                (2 * np.pi / self._lambda) ** 2 * 
                np.sum(self._layersJs * 
                       self._layersWindSpeed ** (5. / 3))
                ) ** (-3. / 5) * u.s

    def outer_scale(self):
        return self._layersL0 * u.m


class MaorySteroScidarProfiles():

    def __init__(self):
        pass

    @staticmethod
    def _retrieveMaoryP():
        rootDir = dataRootDir()
        filename = os.path.join(rootDir,
                                'cn2profiles',
                                'maoryPProfiles.fits')
        return fits.getdata(filename)

    @classmethod
    def _profileMaker(cls, idxJ, idxW):
        rr = cls._retrieveMaoryP()
        h = rr[0] * 1e3
        js = rr[idxJ]
        L0s = np.ones(len(js)) * 30.
        windSpeed = rr[idxW]
        windDirection = np.linspace(0, 360, len(js))
        return Cn2Profile(js, L0s, h, windSpeed, windDirection)

    @classmethod
    def P10(cls):
        return cls._profileMaker(1, 6)

    @classmethod
    def P25(cls):
        return cls._profileMaker(2, 7)

    @classmethod
    def P50(cls):
        return cls._profileMaker(3, 8)

    @classmethod
    def P75(cls):
        return cls._profileMaker(4, 9)

    @classmethod
    def P90(cls):
        return cls._profileMaker(5, 10)


class MaoryStereoScidarProfiles2021():

    L0 = 25

    @staticmethod
    def _restoreMaoryProfiles():
        rootDir = dataRootDir()
        filename = os.path.join(rootDir,
                                'cn2profiles',
                                'referenceProfiles35layers_new.fits')
        return fits.getdata(filename)

    @classmethod
    def _profileMaker(cls, idxJ, idxW):
        rr = cls._restoreMaoryProfiles()
        h = rr[0] * 1e3
        js = rr[idxJ]
        L0s = np.ones(len(js)) * cls.L0
        windSpeed = rr[idxW]
        windDirection = np.linspace(0, 360, len(js))
        return Cn2Profile(js, L0s, h, windSpeed, windDirection)

    @classmethod
    def P10(cls):
        return cls._profileMaker(1, 6)

    @classmethod
    def P25(cls):
        return cls._profileMaker(2, 7)

    @classmethod
    def P50(cls):
        return cls._profileMaker(3, 8)

    @classmethod
    def P75(cls):
        return cls._profileMaker(4, 9)

    @classmethod
    def P90(cls):
        return cls._profileMaker(5, 10)


class EsoEltProfiles():

    L0 = 25

    @staticmethod
    def _restoreProfiles():
        rootDir = dataRootDir()
        filename = os.path.join(rootDir,
                                'cn2profiles',
                                'cn2_eso_258292_2.fits')
        return fits.getdata(filename, header=True)

    @classmethod
    def _profileMaker(cls, keyR0, idxJ, idxWind, L0=None):
        rr, hdr = cls._restoreProfiles()
        r0 = hdr[keyR0]
        h = rr[0]
        js = rr[idxJ] / 100
        windSpeed = rr[idxWind]
        windDirection = np.linspace(0, 360, len(js))
#         windDirection = np.random.uniform(0, 360, len(js))
        if L0 is None:
            L0 = cls.L0
        L0s = np.ones(len(js)) * L0
        return Cn2Profile.from_fractional_j(
            r0, js, L0s, h, windSpeed, windDirection)

    @classmethod
    def Median(cls, *args, **kwargs):
        return cls._profileMaker('R0MEDIAN', 1, 6, *args, **kwargs)

    @classmethod
    def Q1(cls):
        return cls._profileMaker('R0Q1', 2, 7)

    @classmethod
    def Q2(cls):
        return cls._profileMaker('R0Q2', 3, 8)

    @classmethod
    def Q3(cls):
        return cls._profileMaker('R0Q3', 4, 9)

    @classmethod
    def Q4(cls):
        return cls._profileMaker('R0Q4', 5, 10)


class MiscellaneusProfiles():

    @classmethod
    def MaunaKea(cls):
        '''
        From Brent L. Ellerbroek, Francois J. Rigaut,
        "Scaling multiconjugate adaptive optics performance estimates
        to extremely large telescopes,"
        Proc. SPIE 4007, Adaptive Optical Systems Technology,
        (7 July 2000); doi: 10.1117/12.390314
        '''
        r0 = 0.236
        hs = np.array([0.09, 1.826, 2.72, 4.256, 6.269, 8.34,
                       10.546, 12.375, 14.61, 16.471, 17.028]) * 1000
        js = [0.003, 0.136, 0.163, 0.161, 0.167,
              0.234, 0.068, 0.032, 0.023, 0.006, 0.007]
        windSpeed = np.ones(len(js)) * 10.0
        windDirection = np.linspace(0, 360, len(js))
        L0s = np.ones(len(js)) * 25.0
        return Cn2Profile.from_fractional_j(
            r0, js, L0s, hs, windSpeed, windDirection)

    @classmethod
    def LBT(cls):
        '''
        From G. Agapito, C. Arcidiacono, F. Quiros-Pacheco, S. Esposito,
        "Adaptive optics at short wavelengths - Expected performance and sky
        coverage of the FLAO system going toward visible wavelengths",
        doi: 10.1007/s10686-014-9380-7
        '''
        r0 = 0.140
        hs = np.array([0.103, 0.725, 2.637, 11.068]) * 1000
        js = [0.70, 0.06, 0.14, 0.10]
        windSpeed = [2., 4., 6., 25.]
        windDirection = [0., 90., 180., 270.]
        L0s = np.ones(len(js)) * 40.0
        return Cn2Profile.from_fractional_j(
            r0, js, L0s, hs, windSpeed, windDirection)

    @classmethod
    def ERIS(cls):
        '''
        From Doc. No.: VLT-SPE-ESO-11250-4110 
        '''
        r0 = 0.12633
        hs = np.array([0.03, 0.140, 0.281, 0.562, 1.125, 2.250, 4.500,
                       7.750, 11.000, 14.000]) * 1000
        js = [0.59, 0.02, 0.04, 0.06, 0.01, 0.05, 0.09, 0.04, 0.05, 0.05]
        windSpeed = [6.6, 5.9, 5.1, 4.5, 5.1, 8.3, 16.3, 30.2, 34.3, 17.5]
        windDirection = np.linspace(0, 360, len(js))
        L0s = np.ones(len(js)) * 25.
        # L0s = np.ones(len(js)) * 22.
        return Cn2Profile.from_fractional_j(
            r0, js, L0s, hs, windSpeed, windDirection)



import astropy.units as u
import os
import tempfile
import subprocess
from astropy.io import fits


class EsoSkyCalc(object):
    '''
    Interface to `EsoSkyCalc <http://www.eso.org/observing/etc/bin/gen/form?INS.MODE=swspectr+INS.NAME=SKYCALC>`_ 

    Calling sequence::

       sky = EsoSkyCalc(**kwargs)

    the full list of keywords parameters and defaults is in `the help page <http://www.eso.org/observing/etc/doc/skycalc/helpskycalccli.html>`_ 

    Examples
    --------
    Plot sky radiance for airmass=2 and full moon

    >>> sky = EsoSkyCalc(airmass=2, moon_sun_sep=180) 
    >>> plt.semilogy(sky.lam, sky.flux)

    Retrieve dictionary of default values

    >>> defdict= EsoSkyCalc.default_values()
    >>> defdict['observatory'] == '3060m'


    '''

    def __init__(self, **kwargs):
        self._defaults = EsoSkyCalc.default_values()
        for key, value in kwargs.items():
            self._defaults[key] = value
        self._create_input_file()
        self._run_cli()
        self._restore_fits()
        self._u = u.ph / u.s / u.m ** 2 / u.um / u.arcsec ** 2

    def _temp_input_file_name(self):
        return os.path.join(tempfile.gettempdir(), 'eso_sky_calc_input.txt')

    def _temp_output_file_name(self):
        return os.path.join(tempfile.gettempdir(), 'eso_sky_calc_output.fits')

    def _create_input_file(self):
        tmp_file_name = self._temp_input_file_name()
        with open(tmp_file_name, 'w') as the_file:
            for key, value in self._defaults.items():
                the_file.write('%s : %s\n' % (key, value))

    def _run_cli(self):
        res = subprocess.run([
            "skycalc_cli",
            "-i",
            self._temp_input_file_name(),
            "-o",
            self._temp_output_file_name()
            ],
            check=True)
        assert res.returncode == 0

    def _restore_fits(self):
        self._res = fits.getdata(self._temp_output_file_name())
        self._hdr = fits.getheader(self._temp_output_file_name())

    @staticmethod
    def default_values():
        dd = {}
        dd['airmass'] = 1.155
        dd['pwv_mode'] = 'season'
        dd['season'] = 0
        dd['time'] = 0
        dd['pwv'] = 2.5
        dd['msolflux'] = 130.0
        dd['incl_moon'] = 'Y'
        dd['moon_sun_sep'] = 0.0
        dd['moon_target_sep'] = 25.0
        dd['moon_alt'] = 45.0
        dd['moon_earth_dist'] = 1.0
        dd['incl_starlight'] = 'Y'
        dd['incl_zodiacal'] = 'Y'
        dd['ecl_lon'] = 135.0
        dd['ecl_lat'] = 90.0
        dd['incl_loweratm'] = 'Y'
        dd['incl_upperatm'] = 'Y'
        dd['incl_airglow'] = 'Y'
        dd['incl_therm'] = 'N'
        dd['therm_t1'] = 0.0
        dd['therm_e1'] = 0.0
        dd['therm_t2'] = 0.0
        dd['therm_e2'] = 0.0
        dd['therm_t3'] = 0.0
        dd['therm_e3'] = 0.0
        dd['vacair'] = 'vac'
        dd['wmin'] = 300.0
        dd['wmax'] = 5000.0
        dd['wgrid_mode'] = 'fixed_wavelength_step'
        dd['wdelta'] = 10
        dd['wres'] = 20000
        dd['lsf_type'] = 'none'
        dd['lsf_gauss_fwhm'] = 5.0
        dd['lsf_boxcar_fwhm'] = 5.0
        dd['observatory'] = '3060m'
        dd['temp_flag'] = 1
        return dd

    @property
    def lam(self):
        '''
        Wavelength array in astropy units of nm
        '''
        return self._res['lam'] * u.nm

    @property
    def flux(self):
        '''
        Radiance array in ph / s / m^2 / um / u.arcsec^2
        '''
        return self._res['flux'] * self._u

    @property
    def trans(self):
        '''
        Fractional transmission
        '''
        return self._res['trans']

    @property
    def flux_sml(self):
        '''
        Scattered moonlight in ph / s / m^2 / um / u.arcsec^2
        '''
        return self._res['flux_sml'] * self._u

    @property
    def flux_ssl(self):
        '''
        Scattered starlight in ph / s / m^2 / um / u.arcsec^2
        '''
        return self._res['flux_ssl'] * self._u

    @property
    def flux_zl(self):
        '''
        Zodiacal light in ph / s / m^2 / um / u.arcsec^2
        '''
        return self._res['flux_zl'] * self._u

    @property
    def flux_tme(self):
        '''
        Molecular emission of lower atmosphere in ph / s / m^2 / um / u.arcsec^2
        '''
        return self._res['flux_tme'] * self._u

    @property
    def flux_ael(self):
        '''
        Emission lines of upper atmosphere in ph / s / m^2 / um / u.arcsec^2
        '''
        return self._res['flux_ael'] * self._u

    @property
    def flux_arc(self):
        '''
        Airglow / residual continuum in ph / s / m^2 / um / u.arcsec^2
        '''
        return self._res['flux_arc'] * self._u

    @property
    def flux_tie(self):
        '''
        Telescope / Instrument thermal emission in ph / s / m^2 / um / u.arcsec^2
        '''
        return self._res['flux_tie'] * self._u

    @property
    def trans_ma(self):
        '''
        Molecular Absorption
        '''
        return self._res['trans_ma']

    @property
    def trans_o3(self):
        '''
        Ozone UV / optical absorption
        '''
        return self._res['trans_o3']

    @property
    def trans_rs(self):
        '''
        Rayleigh scattering
        '''
        return self._res['trans_rs']

    @property
    def trans_ms(self):
        '''
        Mie scattering
        '''
        return self._res['trans_ms']



import numpy as np
from astropy import units as u
from synphot.spectrum import SpectralElement
from synphot.models import Empirical1D
from synphot.utils import generate_wavelengths, merge_wavelengths
from astropy.io import fits


class Bandpass():
    
    _MIN_LAMBDA_NM = 200
    _MAX_LAMBDA_NM = 10000
    _DELTA_LAMBDA_NM = 10

    @classmethod
    def standard_waveset(cls):
        return generate_wavelengths(
            minwave=cls._MIN_LAMBDA_NM,
            maxwave=cls._MAX_LAMBDA_NM,
            delta=cls._DELTA_LAMBDA_NM,
            log=False, wave_unit=u.nm)[0]

    @classmethod
    def one(cls):
        return cls.flat(1)

    @classmethod
    def zero(cls):
        return cls.flat(0)

    @classmethod
    def flat(cls, amplitude):
        wv = cls.standard_waveset()
        return SpectralElement(
            Empirical1D, points=wv, lookup_table=amplitude * np.ones(wv.shape))

    @classmethod
    def from_array(cls, waveset, values):
        return SpectralElement(Empirical1D, points=waveset, lookup_table=values)

    @classmethod
    def ramp(cls, low_wl, low_ampl, high_wl, high_ampl):
        '''
        T=low_ampl for lambda < low_wl
        T=high_ampl for lambda > high_wl
        T=linear_interpolation(low_ampl, high_ampl) for low_wl < lambda < high_wl
        '''
        l1 = low_wl.to(u.nm).value
        l2 = high_wl.to(u.nm).value
        return SpectralElement(
            Empirical1D,
            points=np.array([cls._MIN_LAMBDA_NM, l1, l2, cls._MAX_LAMBDA_NM]
                            ) * u.nm,
            lookup_table=np.array([low_ampl, low_ampl, high_ampl, high_ampl]))
        
    @classmethod
    def peak(cls, peak_wl, delta_wl, high_ampl, low_ampl):
        '''
        T = high_ampl for lambda = peak_wl
        T = low_ampl for lambda < (peak_wl - delta_wl) and 
            lambda > (peak_wl + delta_wl)
        T = linear_interpolation(low_ampl, high_ampl) for 
            (peak_wl - delta_wl) < lambda < peak_wl and
            peak_wl < lambda < (peak_wl + delta_wl) 
        '''
        l = peak_wl.to(u.nm).value
        dl = delta_wl.to(u.nm).value
        return SpectralElement(
            Empirical1D,
            points=np.array([
                cls._MIN_LAMBDA_NM, l - dl, l, l + dl, cls._MAX_LAMBDA_NM]) * u.nm,
            lookup_table=np.array([
                low_ampl, low_ampl, high_ampl, low_ampl, low_ampl]))
        
    @classmethod
    def top_hat(cls, peak_wl, delta_wl, high_ampl, low_ampl):
        '''
        T = high_ampl for lambda = peak_wl +/- delta_wl
        T = low_ampl elsewhere
        '''
        l = peak_wl.to(u.nm).value
        dl = delta_wl.to(u.nm).value
        return SpectralElement(
            Empirical1D,
            points=np.array([
                cls._MIN_LAMBDA_NM, l - dl, l - dl + 1e-8, l + dl,
                l + dl + 1e-8, cls._MAX_LAMBDA_NM]) * u.nm,
            lookup_table=np.array([
                low_ampl, low_ampl, high_ampl, high_ampl, low_ampl, low_ampl]))

    @classmethod
    def step(cls, wl, low_ampl, high_ampl):
        '''
        T=low_ampl for lambda < wl
        T=high_ampl for lambda > wl
        '''
        return cls.ramp(wl, low_ampl, wl + 1e-8 * u.Angstrom, high_ampl)


class TransmissiveElement():
    '''
    An optical element with transmittance, reflectance and absorptance

    Only 2 of the 3 parameters must be specified, as the sum of the 3 quantities
    must be equal to 1 by energy conservation at every wavelength.

    Absorptance and absorptivity are synonimous. The emissivity is equal to the
    absorptance under the assumptions of Kirchhoff's law of thermal radiation.
    (Absorpance is instead a quantity related to the optical depth).

    When a scalar value is specified, a constant spectrum of the specified
    amplitude is assumed

    Parameters
    ----------
        transmittance: `synphot.SpectralElement` or scalar in [0,1]
            spectral transmittance of the element (in wavelength)

        reflectance: `synphot.SpectralElement` or scalar in [0,1]
            spectral reflectance of the element (in wavelength)

        absorptance: `synphot.SpectralElement` or scalar in [0,1]
            spectral absorptance of the element (in wavelength)
    '''

    def __init__(self, transmittance=None, reflectance=None, absorptance=None):
        self._initialize(transmittance, reflectance, absorptance)

    def _initialize(self, transmittance=None, reflectance=None, absorptance=None):
        t, r, a = self._check_params(transmittance, reflectance, absorptance)
        self._t = t
        self._r = r
        self._a = a

    def _check_params(self, t, r, a):
        if t is None:
            if r is None or a is None:
                raise TypeError("Only one among t,r,a can be None")
            t = self._computeThird(r, a)
            return t, r, a
        if r is None:
            if t is None or a is None:
                raise TypeError("Only one among t,r,a can be None")
            r = self._computeThird(t, a)
            return t, r, a
        if a is None:
            if t is None or r is None:
                raise TypeError("Only one among t,r,a can be None")
            a = self._computeThird(t, r)
            return t, r, a
        raise TypeError("At least one among t,r,a must be None")

    def _computeThird(self, a, b):
        wv = a.waveset
        ones = Bandpass.one()
        c = SpectralElement(
            Empirical1D, points=wv,
            lookup_table=ones(wv) - a(wv) - b(wv))
        if (np.max(c(wv)) > 1.0) or (np.min(c(wv) < 0)):
            raise ValueError("t+r+a=1 cannot be fulfilled")
        return c

    def plot(self, transmittance=True, reflectance=True,
             absorptance=True, wv_unit=None, **kwargs):
        import matplotlib.pyplot as plt
        if wv_unit is None:
            wv = self.waveset
        else:
            wv = self.waveset.to(wv_unit)
        if transmittance:
            plt.plot(wv, self.transmittance(wv), label='t', **kwargs)
        if reflectance:
            plt.plot(wv, self.reflectance(wv), label='r', **kwargs)
        if absorptance:
            plt.plot(wv, self.absorptance(wv), label='a', **kwargs)
        plt.legend()
        plt.grid(True)

    @property
    def waveset(self):
        return self.transmittance.waveset

    @property
    def transmittance(self):
        return self._t

    @property
    def absorptance(self):
        return self._a

    @property
    def reflectance(self):
        return self._r

    @property
    def emissivity(self):
        return self.absorptance
    
    def _compute_average_in_band(self, spectral_el, wv_band, atol):
        id_wv_min = np.where(np.isclose(self.waveset, wv_band[0], atol=atol[0]))[0]
        wv_min = self.waveset[id_wv_min]
        if len(wv_min) != 1:
            raise ValueError(f'Wavelength found: {wv_min}. Change atol.')
        
        id_wv_max = np.where(np.isclose(self.waveset, wv_band[1], atol=atol[1]))[0]
        wv_max = self.waveset[id_wv_max]
        if len(wv_max) != 1:
            raise ValueError(f'Wavelength found: {wv_max}. Change atol.')
    
        return wv_min, wv_max, np.mean(spectral_el(self.waveset)[id_wv_min[0]:id_wv_max[0]+1]) 

    def transmittance_in_band(self, wv_band, atol=(1, 1)):
        wv_min, wv_max, t_mean = self._compute_average_in_band(self.transmittance, wv_band, atol)
        print(f'Average transmittance in band {(wv_min, wv_max)}: {t_mean}')
        return t_mean
    
    def reflectance_in_band(self, wv_band, atol=(1, 1)):
        wv_min, wv_max, r_mean = self._compute_average_in_band(self.reflectance, wv_band, atol)
        print(f'Average reflectance in band {(wv_min, wv_max)}: {r_mean}')
        return r_mean
    
    def emissivity_in_band(self, wv_band, atol=(1, 1)):
        wv_min, wv_max, e_mean = self._compute_average_in_band(self.emissivity, wv_band, atol)
        print(f'Average emissivity in band {(wv_min, wv_max)}: {e_mean}')
        return e_mean

    def _insert_spectral_element(self, spectral_element, new_values, wavelengths):
        if new_values is None:
            return None
        ws = spectral_element.waveset
        ii = np.searchsorted(ws, wavelengths)
        new_waveset = np.insert(ws, ii, wavelengths)
        
        r_arr = spectral_element(ws)
        new_r = np.insert(r_arr, ii, new_values)
        return SpectralElement(Empirical1D,
                            points=new_waveset,
                            lookup_table=new_r)
       
    def add_spectral_points(self, wavelengths, transmittance=None,
                            reflectance=None, absorptance=None):
        new_t = self._insert_spectral_element(self.transmittance, transmittance,
                                               wavelengths)
        new_r = self._insert_spectral_element(self.reflectance, reflectance,
                                              wavelengths)
        new_a = self._insert_spectral_element(self.absorptance, absorptance,
                                              wavelengths)

        self._initialize(transmittance=new_t, reflectance=new_r,
                         absorptance=new_a)
        
    def to_dat(self, filepath, data_type):
        if data_type == 'transmittance':
            data = self.transmittance(self.waveset).value
            title = np.array([['Wavelength [um]', 'Transmittance']])
        elif data_type == 'reflectance':
            data = self.reflectance(self.waveset).value
            title = np.array([['Wavelength [um]', 'Reflectance']])
        elif data_type == 'absorptance':
            data = self.absorptance(self.waveset).value
            title = np.array([['Wavelength [um]', 'Absorptance']])
        to_write = np.stack((
            self.waveset.to(u.um).value,
            data), axis=1)
        with open(filepath, 'a') as datafile:
            np.savetxt(datafile,
                       np.vstack((title, to_write)),
                       fmt=['%-30s', '%-30s'])

    def to_fits(self, filename, **kwargs):
        wv = self.waveset
        hdu1 = fits.PrimaryHDU(wv.to(u.nm).value)
        hdu2 = fits.ImageHDU(self.transmittance(wv).value)
        hdu3 = fits.ImageHDU(self.absorptance(wv).value)
        hdul = fits.HDUList([hdu1, hdu2, hdu3])
        hdul.writeto(filename, **kwargs)

    @staticmethod
    def from_fits(filename):
        wv = fits.getdata(filename, 0) * u.nm
        t = SpectralElement(Empirical1D, points=wv,
                            lookup_table=fits.getdata(filename, 1))
        a = SpectralElement(Empirical1D, points=wv,
                            lookup_table=fits.getdata(filename, 2))
        return TransmissiveElement(transmittance=t, absorptance=a)


class Direction(object):
    TRANSMISSION = 'transmission'
    REFLECTION = 'reflection'


class TransmissiveSystem():

    def __init__(self):
        self._elements = []
        self._waveset = np.array([1000]) * u.nm

    def add(self, transmissive_element, direction):
        self._elements.append(
            {'element': transmissive_element,
             # 'name': f'{transmissive_element = }'.split()[0],
             'direction': direction})
        self._waveset = merge_wavelengths(
            self._waveset.to(u.nm).value,
            transmissive_element.waveset.to(u.nm).value) * u.nm

    def remove(self, element_index):
        del self._elements[element_index]

    @property
    def transmittance(self):
        return self._compute_transmittance()
    
    def transmittance_from_to(self, from_element=0, to_element=None):
        return self._compute_transmittance(from_element, to_element)

    def _compute_transmittance(self, from_element=0, to_element=None):
        t = 1
        for el in self._elements[from_element:to_element]:
            direction = el['direction']
            if direction == Direction.TRANSMISSION:
                t *= el['element'].transmittance
            elif direction == Direction.REFLECTION:
                t *= el['element'].reflectance
            else:
                raise ValueError('Unknown propagation direction %s' % direction)
        return t

    @property
    def emissivity(self):
        return self._compute_absorptance()

    def _compute_absorptance(self):
        a = Bandpass.zero()
        for i, el in enumerate(self._elements[:-1]):
            partial_absorptance = \
                el['element'].absorptance * self._compute_transmittance(
                    from_element=i + 1)
            a = self._sum_absorptance(a, partial_absorptance)
        a = self._sum_absorptance(a, self._elements[-1]['element'].absorptance)
        return a

    def _sum_absorptance(self, a, b):
        return Bandpass.from_array(self._waveset,
                                   a(self._waveset) + b(self._waveset))

    @property
    def number_of_transmissive_elements(self):
        return len(self._elements)
    
    def as_transmissive_element(self):
        return TransmissiveElement(transmittance=self.transmittance,
                                   absorptance=self.emissivity)

    def plot(self, **kwargs):
        self.as_transmissive_element().plot(reflectance=False, **kwargs)
        



from arte.types.mask import CircularMask


import numpy as np


class KarhunenLoeveGenerator(object):
    """
    Generator of KL polynomials departing from a covariance matrix and a user
    defined base of polynomials.

    The user can specify the diameter of the pupil `pupil` or
    specify a `~arte.types.mask.CircularMask`

    In the case a scalar real value is passed as `pupil` argument,
    The shape of the returned array cannot be specified: it is a
    square array of size ceil(pupil_diameter).
    The center of the unit disk cannot be specified and it is the central pixel
    of the array (if the array size is odd) or the corner of the
    4 central pixels (if the array size is even).


    In the case a `~arte.types.mask.CircularMask` is passed as `pupil` argument,
    the mask is used as unit disk over which the polynomial are computed. Non-integer
    center coordinates and radius are properly managed.



    Parameters
    ----------
        pupil: real or `~arte.types.mask.CircularMask`
            If a scalar value, the argument is used as pupil diameter in pixels.
            If a `~arte.types.mask.CircularMask`, the argument is used as mask
            representing the unit disk.
        covariance_matrix:  `~numpy.array` of shape (nModes, nModes) representing
                            the phenomenon covariance matrix


    Examples
    --------

    Create a KL polynomial using Zernike description of Kolmogorov turbulence
    >>> from arte.atmo.utils import getFullKolmogorovCovarianceMatrix
    >>> from arte.utils.zernike_generator import ZernikeGenerator
    >>> from arte.utils.kl_generator import KLGenerator
    >>> from arte.types.mask import CircularMask
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt

    >>> mask = CircularMask((128,128), 64, (64,64))
    >>> zz = ZernikeGenerator(mask)
    >>> nmodes=100
    >>> kolm_covar = getFullKolmogorovCovarianceMatrix(nmodes)
    >>> zbase = np.rollaxis(np.ma.masked_array([zz.getZernike(n)
                                            for n in range(2,nmodes+2)]),0,3)
    >>> generator = KLGenerator(mask, kolm_covar)
    >>> generator.generateFromBase(zbase)
    >>> kl =generator.getKL(2)
    >>> plt.imshow(kl)


    """

    def __init__(self, pupil, covariance_matrix):
        if isinstance(pupil, CircularMask):
            self._radius = pupil.radius()
            self._shape = pupil.shape()
            self._center = pupil.center()
            self._boolean_mask = pupil.mask()
        else:
            self._radius = pupil / 2
            sz = np.ceil(pupil)
            self._shape = (sz, sz)
            self._center = np.ones(2) * (sz / 2)
            cm = CircularMask(
                self._shape, maskCenter=self._center, maskRadius=self._radius
            )
            self._boolean_mask = cm.mask()

        self._covariance_matrix = covariance_matrix
        self._rhoMap, self._thetaMap = self._polar_array()
        self._dx = None
        self._dy = None
        self._dictCache = {}
        self._dictDxCache = {}
        self._dictDyCache = {}
        self._nmodes = covariance_matrix.shape[0]
        self._inpupil = pupil
        self._klBase = None

    def radius(self):
        """
        Radius of the unit disk

        Returns
        -------
        radius: real
            unit disk radius in pixel

        """
        return self._radius

    def center(self):
        """
        Y, X coordinates of the unit disk center

        Returns
        -------
        center: `~numpy.array` of shape (2,)
            Y, X coordinate of the unit disk center in the array reference system

        """
        return self._center

    def first_mode(self):
        '''
        First mode index
        '''
        return 0
    
    def _derivativeCoeffX(self, index):
        raise Exception("Not available here")

    def _derivativeCoeffY(self, index):
        raise Exception("Not available here")

    @classmethod
    def degree(cls, index):
        raise Exception("Not available here")

    def _rnm(self, radialDegree, azimuthalFrequency, rhoArray):
        raise Exception("Not available here")

    def _polar(self, index, rhoArray, thetaArray):
        raise Exception("Not available here")

    def cartesian_coordinates(self):
        """
        Return X, Y maps of cartesian coordinates

        The KL polynomials and derivatives are evaluated over a grid of
        points, whose coordinates are accessible through `cartesian_coordinates`

        Returns
        -------
        X,Y: `~numpy.array`
            coordinates of points of the unit disk where the polynomials
            are evaluated

        Examples
        --------

        Create KL polynomials on a unit disk defined over
        4 pixels. The returned arrays have (4,4) shape (corresponding
        to the coords [-0.75, -0.25, 0.25, 0.75])

        >>> zg = KLGenerator(4)
        >>> x, y = zg.cartesian_coordinates()
        >>> x[0]
        array([-0.75, -0.25,  0.25,  0.75])

        As above with 3 pixels

        >>> zg = KLGenerator(3)
        >>> x, y = zg.cartesian_coordinates()
        >>> x[0]
        array([-0.666, 0,  0.666])

        In case of non-integer diameter, the array size is rounded up to the
        next integer

        >>> zg = KLGenerator(2.5)
        >>> x, y = zg.cartesian_coordinates()
        >>> x[0]
        array([-0.8, 0,  0.8])

        """
        nPxY = self._shape[0]
        nPxX = self._shape[1]
        c = np.array(self.center())
        cc = np.expand_dims(c, axis=(1, 2))
        Y, X = (
            np.mgrid[0.5 : nPxY + 0.5 : 1, 0.5 : nPxX + 0.5 : 1] - cc
        ) / self.radius()

        return X, Y

    def _polar_array(self):
        X, Y = self.cartesian_coordinates()
        r = np.sqrt(X**2 + Y**2)
        th = np.arctan2(Y, X)
        return r, th

    def generateFromBase(self, inBase):
        # Generate the KL base from a base of polynomials.
        # The input base is a cube of polynomials in the form (nPx, nPx, nmodes)
        # where nPx is the size of the square array where the polynomials are defined
        # and nmodes is the number of modes in the base.
        # The base is used to compute the Karhunen-Loeve modes of the phenomena
        # defined by the covariance matrix.
        # The base is internally stored and can be accessed through the getCube
        # method.
        # The base is computed as the dot product of the input base and the
        # eigenvectors of the covariance matrix.
        # The input base is used to compute the Karhunen-Loeve modes, which are
        # stored in the _klBase attribute.

        shapex, shapey, nmodes = inBase.shape
        if nmodes != self._nmodes:
            raise ValueError(
                "The number of modes in the base is different from the number of modes in the covariance matrix"
            )

        self._inBase = inBase
        # zcube_masked = zcube.compressed().reshape(np.size(zcube.compressed())//nzern,nzern)
        # Compute the eigenvalues and eigenvectors
        U, S, Vt = np.linalg.svd(self._covariance_matrix)

        # The eigenvectors are the Karhunen-Loeve modes
        klcube = inBase.dot(Vt.T)

        # Remove piston from each mode
        klcube = klcube - np.mean(klcube, axis=(0, 1))
        self._klBase = klcube
        # refill the _dictCache with the new base
        self._dictCache = {i: klcube[:, :, i] for i in range(0, nmodes)}
        self._klBase = np.rollaxis(
            np.ma.masked_array([self.getKL(n) for n in range(self._nmodes)]), 0, 3
        )

    def getCube(self):
        # Return the base of KL polynomials
        if self._klBase is None:
            raise ValueError("KL base not generated")
        else:
            return self._klBase

    def getKLDict(self, indexVector):
        # Return a dictionary of KL polynomials indexed by the input vector
        ret = {}
        for index in indexVector:
            ret[index] = self.getKL(index)
        return ret

    def getModesDict(self, indexVector):
        return self.getKLDict(indexVector)

    def getKL(self, index):
        # Return the KL polynomial corresponding to the input index
        if not self._is_integer_num(index):
            raise ValueError("Invalid KL index %s not integer" % index)
        if index < 0:
            raise ValueError("Invalid KL index %d less than 0" % index)
        if index not in list(self._dictCache.keys()):
            raise ValueError("Invalid KL index %d, out of base" % index)
        return self._dictCache[index]

    @staticmethod
    def _is_integer_num(n):
        if isinstance(n, (int, np.integer)):
            return True
        if isinstance(n, float):
            return n.is_integer()
        return False

    def __getitem__(self, index):
        return self.getKL(index)

    def getDerivativeXDict(self, indexVector):
        raise Exception("Not implemented yet")

    def getDerivativeYDict(self, indexVector):
        raise Exception("Not implemented yet")

    def getDerivativeX(self, index):
        raise Exception("Not implemented yet")
        # if index not in self._dictDxCache:
        #    self._dictDxCache[index] = self._computeDerivativeX(index)
        # return self._dictDxCache[index]

    def _computeDerivativeX(self, index):
        pass

    def getDerivativeY(self, index):
        raise Exception("Not implemented yet")

    def _computeDerivativeY(self, index):
        pass




import itertools

def chunk_iterator(n, iterable):
    '''
    From: https://stackoverflow.com/questions/8991506/iterate-an-iterator-by-chunks-of-n-in-python
    - without permission

    Splits an iterable in chunks of N elements each (but for the last one,
    which might be shorter if needed). Returns an iterator for each chunk.
    Example:
    >>> for a in chunk_iterator(3,range(10)):
    ...  print(list(a))
    ... 
    [0, 1, 2]
    [3, 4, 5]
    [6, 7, 8]
    [9]
    '''
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)





from functools import cached_property
import numpy as np
from arte.dataelab.base_data import BaseData

class Base2dMap(BaseData):
    '''
    2d map. Initialize with a 2d map
    where valid elements are 1 and not used elements are 0.
    '''
    def __init__(self, map2d):
        super().__init__(map2d)

    @staticmethod
    def from_idx1d(shape, idx1d):
        f2d = np.zeros(shape, dtype=bool)
        f2d.flat[idx1d] = True
        return Base2dMap(f2d)

    @cached_property
    def nvalid(self):
        '''Number of valid elements'''
        nonzero = np.nonzero(self.get_data())
        return len(nonzero[0])

    def as_mask(self):
        ''''
        Return a mask suitable to be used in a masked array
        '''
        return (1 - self.get_data()).astype(bool)

    def remap_image(self, data):
        '''
        Remap a 2d time series data into a 3d cube, where each
        cube slice is a 2d image. If a 1d array is passed,
        the resulting 3d array will have a first dimension with length 1.
        
        Parameters
        ----------
        data: ndarray
            2d numpy array [time, values] or 1d numpy array [values]
        
        Result
        ------
        frame: ndarray
            3d numpy array [time, x, y]
        '''
        if len(data.shape) == 1:
            data = np.expand_dims(data, axis=0)

        nonzero = np.nonzero(self.get_data())
        frame = np.zeros((len(data), self.shape[0], self.shape[1]), dtype=data.dtype)
        if len(nonzero[0]) != data.shape[1]:
            raise ValueError(f'Cannot build display: data shape and subap map shape do not match: %s and %d elements' % (data.shape, len(nonzero[0])))
        frame[:, nonzero[0], nonzero[1]] = data
        return frame



import math


def round_up_to_even(f):
    return math.ceil(f / 2.) * 2


def kroneckerDelta(m, n):
    if m == n:
        delta = 1
    else:
        delta = 0
    return delta


def besselFirstKind(order, arg):
    from scipy import special
    return special.jv(order, arg)


def besselFirstKindOnGPU(order, arg):
    import cupy as cp
    from cupyx.scipy import special
    if order == 0:
        bess = special.j0(arg)
    elif order == 1:
        bess = special.j1(arg)
    elif order >= 2:
        bess = 2 * (float(order) - 1) * besselFirstKindOnGPU(int(order) - 1, arg) / cp.real(arg) - besselFirstKindOnGPU(
            int(order) - 2, arg)
        if not cp.all(arg):
            zero_idx = cp.where(arg == 0)
            bess[zero_idx] = 0
    return bess



'''
Axis handler class
'''

class AxisHandler():

    def __init__(self, axes=None):
        '''
        Axes transpositions handlers

        Parameters
        ----------
        axes: sequence of str, optional
            Initial axis order. If None, no axis transpositions will be possible.
        '''
        if axes is None:
            self._axes = ()
        elif isinstance(axes, str):
            self._axes = (axes,)
        else:
            self._axes = tuple(axes)

    def transpose(self, data, axes=None):
        '''
        Parameters
        ----------
        data: ndarray
            data to transpose
        axes: sequence of str, optional
            axis order to apply. If None, no transposition is done.
        '''
        if axes is not None:
            if type(axes) is str:
                axes = [axes]
            # Use an explicit for loop in order to capture an erroneus axis name
            idx = []
            try:
                for ax in axes:
                    idx.append(self._axes.index(ax))
            except ValueError:
                raise ValueError(f'Axis {ax} not found')
            data = data.transpose(idx)
        return data

    def axes(self):
        return self._axes



import numpy as np


def interface_glass_to_glass(n1, n2):
    '''
    Compute reflectance of a glass, considering the refractive index of the
    medium the light is coming from (n1) and the refractive index of the glass
    (n2)
    '''
    return (n1 - n2) ** 2 / (n1 + n2) ** 2


def external_transmittance_calculator(l1, l2, t1, a):
    '''
    Use external transmittance data of a glass with thickness l2 to compute
    external transmittance of a same glass but with different thickness l1.
    The computation is based on the equation for the external transmittance:
    
    T = (1 - R**2) * exp(-a * l)
    
    where R is the reflectance, a is the attenuation coefficient and l is the
    thickness of the glass. If we consider two different values of thickness, 
    thus of transmittance, we can compute the unknown transmittance T2 as:
    
    T2 = T1 *  exp(-a * (l2 - l1))

    '''
    t2 = t1 * np.exp(-a * (l2 - l1))
    return t2


def internal_transmittance_calculator(l1, l2, t1):
    '''
    Use internal transmittance data t1 of a glass with thickness l1 to compute 
    internal transmittance t2 of a same glass but with different thickness l2.
    The transmittance is computed with the following equation:
    
        t2 = t1**(l2/l1)
    '''
    t2 = t1 ** (l2 / l1)
    return t2


def attenuation_coefficient_calculator(l1, l2, t1, t2):
    '''
    Compute attenuation coefficient of a glass from external transmittance data
    of two different values of thickness.
    The computation is based on the equation for the external transmittance
    of a glass:
    
    T = (1 - R**2) * exp(-a * l)
    
    where R is the reflectance, a is the attenuation coefficient and l is the
    thickness of the glass. If we consider two different values of thickness, 
    thus of transmittance, we can compute the ratio between the transmittances:
    
    T1 / T2 = exp(-a * (l1 - l2))
    
    and from this equation we can derive the attenuation coefficient as:
    
    a = (lnT2 - lnT1) / (l1 - l2)
    '''
    a = (np.log(t2) - np.log(t1)) / (l1 - l2)
    return a



import numpy as np
from astropy import units as u
from astropy.units.quantity import Quantity


def wavefront_rms_2_strehl_ratio(wavefrontRms, wavelength):
    """
    Compute Strehl ratio from wavefront aberration according to
    Marechal approximation

    Parameters
    ----------
    wavefrontRms : `astropy.units.quantity.Quantity` or float
        Wavefront aberration rms; if float units must be in meter

    wavelength : `astropy.units.quantity.Quantity` or float
        Wavelength to compute Strehl Ratio; if float units must be in meter


    Returns
    -------
    results : float
        Strehl ratio
    """
    if not isinstance(wavefrontRms, Quantity):
        wavefrontRms = wavefrontRms * u.m
    if not isinstance(wavelength, Quantity):
        wavelength = wavelength * u.m

    return np.exp(-(wavefrontRms / wavelength * 2 * np.pi) ** 2).value


def strehl_ratio_2_wavefront_rms(strehlRatio, wavelength):
    """
    Compute wavefront aberration from Strehl ratio according to
    Marechal approximation

    Parameters
    ----------
    strehlRatio : float
        Strehl ratio (0.0-1.0)

    wavelength : `astropy.units.quantity.Quantity` or float
        Wavelength to compute Strehl Ratio; if float units must be in meter


    Returns
    -------
    wavefrontRms : `astropy.units.quantity.Quantity`
        Wavefront aberration rms

    """
    if not isinstance(wavelength, Quantity):
        wavelength = wavelength * u.m

    return wavelength / (2 * np.pi) * np.sqrt(-1 * np.log(strehlRatio))


def scale_strehl_ratio(strehlRatio, fromWavelength, toWavelength):
    """
    Scale Strehl ratio from one wavelength to another one according to
    Marechal approximation

    Parameters
    ----------
    strehlRatio : float
        Strehl ratio (0.0-1.0)

    fromWavelength : `astropy.units.quantity.Quantity` or float
        Wavelength Strehl Ratio is given at; if float units must be in meter

    toWavelength : `astropy.units.quantity.Quantity` or float
        Wavelength to compute Strehl Ratio at; if float units must be in meter


    Returns
    -------
    strehlRatio : float
        Strehl Ratio computed at toWavelength
    """

    if isinstance(fromWavelength, Quantity):
        fromWavelength = fromWavelength.to(u.m).value
    if isinstance(toWavelength, Quantity):
        toWavelength = toWavelength.to(u.m).value

    return strehlRatio ** (fromWavelength / toWavelength) ** 2



from astropy import units
from arte.utils.constants import Constants
from scipy.special import gamma
from arte.utils.zernike_generator import ZernikeGenerator
import numpy as np


class Seeing(object):

    def __init__(self, seeingAt500nm):
        self._seeingAt500nm = seeingAt500nm
        self._r0At500nm = 0.98 * 500 * units.nm / (
            seeingAt500nm * Constants.ARCSEC2RAD / units.arcsec)

    def to_r0(self, wavelength=500 * units.nm):
        return self._r0At500nm * (wavelength / (500 * units.nm)) ** (6. / 5)

def getKolmogorovCovarianceMatrix(j1, j2):
    n1, m1 = ZernikeGenerator.degree(j1)
    n2, m2 = ZernikeGenerator.degree(j2)

    # Check if m1 is not equal to m2 or if the sum of j1 and j2 is odd and m1 is not zero
    if m1 != m2 or ((not is_even(j1 + j2)) and m1 != 0):
        return 0.0

    # Calculate npn and nmn
    npn = (n1 + n2) // 2 - 1
    nmn = abs(n1 - n2) // 2

    # Calculate the result using the constant from Roddier 1990, divided by 2*pi to convert rad to waves
    result = 0.057494899 * (-1)**((n1 + n2 - 2 * m1) // 2) * np.sqrt((n1 + 1) * (n2 + 1))
    result *= gamma(1.0 / 6.0) / (gamma(17.0 / 6.0))**2 / gamma(29.0 / 6.0)
    c1 = 1.0
    c2 = 1.0

    # Calculate c1 and c2
    if npn > 0:
        for i in range(npn):
            c1 *= (1.0 / 6.0 + i) / (29.0 / 6.0 + i)
    if nmn > 0:
        for i in range(nmn):
            c2 *= (-11.0 / 6.0 + i) / (17.0 / 6.0 + i)

    return result * c1 * c2 * (-1)**nmn
    
def getFullKolmogorovCovarianceMatrix(nModes):
    covar = np.zeros((nModes, nModes))
    for i in range(2, nModes + 2):
        for j in range(2, nModes + 2):
            covar[i - 2, j - 2] = getKolmogorovCovarianceMatrix(i, j)
    return covar

# Helper function to check if a number is even
def is_even(num):
    return num % 2 == 0




'''
@author: giuliacarla
'''

import numpy as np
from scipy.special import jv
import astropy.units as u


class VonKarmanPsd():
    '''
    This class computes the spatial Power Spectral Density (PSD) of turbulent
    phase assuming the Von Karman spectrum.
    The PSD is obtained from the following expression:

    .. math::

        PSD(f,h) = c * r_0(h)^{-5/3} * ( f^2+ \\frac{1}{L_0^2})^{-11/6}

        c = ( \\frac{24}{5} \Gamma(6/5) )^{5/6}  \\frac{\Gamma(11/6)^2}{2 \pi^{11/3} } = 0.0228955


    Parameters
    ----------
    fried_param: float
        Fried parameter characterizing the atmospheric turbulence [m].
    outer_scale: float
        Outer scale of the atmospheric turbulence [m].

    Example
    -------
    Compute the total variance of Kolmogorov over a 10m telescope and 
    compare variance [rad2] with Noll('76): :math:`\Delta_1= 1.029 (D/r_0)^{5/3}`

    >>> R = 5
    >>> r0 = 0.1
    >>> L0 = np.inf
    >>> psd = von_karman_psd.VonKarmanPsd(r0, L0)
    >>> freqs = np.logspace(-8, 4, 1000)
    >>> bess = scipy.special.jv(1, 2*np.pi*R*freqs)
    >>> psdPistonRem = psd.spatial_psd(freqs) * (1 - (bess/(np.pi*R*freqs))**2)
    >>> varInRad2 = np.trapz(psdPistonRem*2*np.pi*freqs, freqs)
    >>> varInRad2Noll = 1.029*(2*R/r0)**(5./3)
    >>> print("%g %g" % (varInRad2, varInRad2Noll))
    2214.36 2216.91

    or use shortcut function von_karman_psd.rms()
    '''
    NUM_CONST = 0.02289558710855519

    def __init__(self, fried_param, outer_scale):
        self._r0 = fried_param
        self._L0 = outer_scale
        self._checkInput()

    def _checkInput(self):
        if np.isscalar(self._r0):
            assert np.isscalar(self._L0), 'L0 must be scalar like r0'
        else:
            assert len(self._r0) == len(self._L0), \
                'len of L0 and r0 differ: %d vs %d' % (
                    len(self._L0), len(self._r0))

    def _computeVonKarmanPsd(self, freqs):
        if type(self._r0) == np.ndarray:
            self._psd = np.array([
                self.NUM_CONST * self._r0[i] ** (-5. / 3) *
                (freqs ** 2 + 1 / self._L0[i] ** 2) ** (-11. / 6)
                for i in range(self._r0.shape[0])])
        else:
            self._psd = self.NUM_CONST * self._r0 ** (-5. / 3) * \
                (freqs ** 2 + 1 / self._L0 ** 2) ** (-11. / 6)

    def spatial_psd(self, freqs):
        '''
        Spatial Power Spectral Density of Von Karman turbulence

        Parameters
        ----------
        freqs: :class:`~numpy:numpy.ndarray`
            Spatial frequencies vector[m^-1].

        Returns
        -------
        psd: :class:`~numpy:numpy.ndarray`
            power spectral density computed at the specified frequencies
        '''
        self._computeVonKarmanPsd(freqs)
        return self._psd

    def plot_von_karman_psd_vs_frequency(self, freqs, idx=None):
        import matplotlib.pyplot as plt
        psd = self.spatial_psd(freqs)
        if idx is None:
            plt.loglog(freqs, psd)
        else:
            plt.loglog(freqs, psd[idx])
        plt.xlabel('Frequency [m$^{-1}$]')
        plt.ylabel('PSD [rad$^{2}$]')


def rms(diameter: u.m,
        wavelength: u.nm,
        fried_param: u.m,
        outer_scale: u.m,
        freqs=None):
    '''
    Von Karman wavefront rms value over a circular aperture

    Parameters
    ----------
    diameter: `~astropy.units.quantity.Quantity` equivalent to meter
        Aperture diameter

    wavelength: `~astropy.units.quantity.Quantity` equivalent to nanometer
        wavelength

    fried_param: `~astropy.units.quantity.Quantity` equivalent to meter
        Fried parameter r0 defined at the specified wavelength

    outer_scale: `~astropy.units.quantity.Quantity` equivalent to meter
        Outer scale L0. Use np.inf for Kolmogorov spectrum

    Other Parameters
    ----------
    freqs:  array of `~astropy.units.quantity.Quantity` equivalent to 1/meter
        spatial frequencies array. Default logspace(-8, 4, 1000) m^-1

    Returns
    -------
    rms: `~astropy.units.quantity.Quantity` equivalent to nm
        wavefront rms for the specified von Karman turbulence
    '''
    R = 0.5 * diameter.to(u.m).value
    wl = wavelength.to(u.nm).value
    r0 = fried_param.to(u.m).value
    L0 = outer_scale.to(u.m).value
    psd = VonKarmanPsd(r0, L0)
    if freqs is None:
        freqs = np.logspace(-8, 4, 1000) / u.m
    freqs = freqs.to(1 / u.m).value
    bess = jv(1, 2 * np.pi * R * freqs)
    psdTotal = psd.spatial_psd(freqs)
    psdPistonRem = psdTotal * (1 - (bess / (np.pi * R * freqs)) ** 2)
    varInRad2 = np.trapz(psdPistonRem * 2 * np.pi * freqs, freqs)
    return np.sqrt(varInRad2) * wl / 2 / np.pi * u.nm




'''
@author: giuliacarla
'''


import numpy as np
import astropy.units as u
from astropy.units.quantity import Quantity


class GuideSource():
    '''
    This class defines the geometry of the guide source of interest.

    Parameters
    ----------
    polar_coords: tuple of two floats
        Source polar coordinates (rho, theta) in [arcsec, degrees].
        theta is positive if generated by a counterclockwise rotation.

    height: float
        Source height in [meters].
    '''

    def __init__(self, polar_coords, height):
        self._rho = polar_coords[0]
        self._theta = polar_coords[1]
        self._z = height

    @staticmethod
    def fromPolarToCartesian(rho, theta, z):
        x = rho * np.cos(theta)
        y = rho * np.sin(theta)
        return x, y, z

    def getSourceCartesianCoords(self):
        if (isinstance(self._rho, Quantity) and (self._theta, Quantity)
                and (self._z, Quantity)):
            rho = self._rho.to(u.arcsec)
            theta = self._theta.to(u.deg)
            z = self._z.to(u.m)
            x, y, z = self.fromPolarToCartesian(
                rho, theta.to(u.rad), z)
            return [x, y, z]
        else:
            x, y, z = self.fromPolarToCartesian(
                self._rho, np.deg2rad(self._theta), self._z)
            return [x * u.arcsec, y * u.arcsec, z * u.m]

    def getSourcePolarCoords(self):
        if (isinstance(self._rho, Quantity) and (self._theta, Quantity)
                and (self._z, Quantity)):
            return [self._rho.to(u.arcsec),
                    self._theta.to(u.deg),
                    self._z.to(u.m)]
        else:
            return [self._rho * u.arcsec, self._theta * u.deg, self._z * u.m]




import astropy.units as u
from arte.utils.unit_checker import make_sure_its_a, separate_value_and_unit


class UnitHandler():
    def __init__(self, wanted_unit):
        self._wanted_unit = wanted_unit
        self._actual_unit = None  # Discovered later
        self._actual_unit_name = None
        self._force = False

    def actual_unit(self):
        '''Data unit as an astropy unit, or None'''
        return self._actual_unit

    def actual_unit_name(self):
        '''Data unit string, or None if not set'''
        return self._actual_unit_name

    def set_force(self, force_flag):
        '''Set unit coercion flag'''
        self._force = force_flag

    def apply_unit(self, data):
        '''Appy to *data* our wanted unit, if any, converting if necessary

        If the force flag is not set, unit will be converted, throwing an
        exception is conversion if impossible.
        If the force flag is set, the old unit is stripped and the new one set.

        If not wanted unit has been set, data is unchanged, including its unit, if any.
        '''
        if self._wanted_unit is not None:
            if self._force:
                value, _ = separate_value_and_unit(data)
                newdata = value * self._wanted_unit
            else:
                newdata =  make_sure_its_a(self._wanted_unit, data, copy=False)
        else:
            newdata = data

        # Cache unit if applied
        if isinstance(newdata, u.Quantity):
            self._actual_unit = newdata.unit
            self._actual_unit_name = newdata.unit.to_string() or 'dimensionless'
        else:
            self._actual_unit = None
            self._actual_unit_name = None

        return newdata




import sys

def yes_no(question, default="yes"):
    """
    From: http://code.activestate.com/recipes/577058/
    MIT license

    Ask a yes/no question via input() and return their answer.
    
    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is one of "yes" or "no".
    """
    valid = {"yes":"yes",   "y":"yes",  "ye":"yes",
             "no":"no",     "n":"no"}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while 1:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return default
        elif choice in valid.keys():
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "\
                             "(or 'y' or 'n').\n")






from abc import abstractmethod
from functools import wraps
import abc
import threading
import traceback

from concurrent import futures

from arte.utils.decorator import override, synchronized
from arte.utils.logger import Logger


class Task(object, metaclass=abc.ABCMeta):

    @abstractmethod
    def perform(self):
        assert False, "abstract base clase"


    def name(self):
        return "unspecified"


    def __repr__(self):
        return "Task(%r)" % self.name()


class FunctionTaskException(Exception):
    pass


class FunctionTask(Task):

    NO_RESULT= object()

    def __init__(self, func, *args, **kwds):
        self._func= func
        self._args= args
        self._kwds= kwds

        self._result= self.NO_RESULT
        self._exception= None
        self._name= None


    @override
    def perform(self):
        try:
            self._result= self._func(*self._args, **self._kwds)
        except Exception as e:
            self._exception= e
            raise


    @override
    def name(self):
        if self._name:
            return self._name

        return "%s" % self._func


    def __repr__(self):
        return "FunctionTask(%r)" % self.name()


    def withName(self, name):
        self._name= name
        return self


    def getResult(self):
        if self._result is self.NO_RESULT:
            raise FunctionTaskException(
                "No result available for %s" % self.name())
        return self._result


    def getException(self):
        return self._exception


    def reraiseExceptionIfAny(self):
        if self._exception:
            raise self._exception


class TaskControl(object, metaclass=abc.ABCMeta):
    @abstractmethod
    def isDone(self):
        assert False, "abstract base class"


    @abstractmethod
    def isRunning(self):
        assert False, "abstract base class"


    @abstractmethod
    def waitForCompletion(self, timeoutSec):
        assert False, "abstract base class"



class TaskTimeoutError(Exception):

    def __init__(self, msg):
        Exception.__init__(self, msg)


class FutureTaskControl(TaskControl):

    def __init__(self, future, taskName, logger):
        TaskControl.__init__(self)
        assert taskName is not None
        self._future= future
        self._taskName= taskName
        self._logger= logger


    @override
    def isDone(self):
        return self._future.done()


    @override
    def isRunning(self):
        return self._future.running()


    def _printStackTraceInCaseOfError(self):
        try:
            self._future.result()
        except Exception as e:
            self._logger.error("Task '%s' failed: %s" % (
                               self._taskName, str(e)))
            traceback.print_exc()
            self._logger.debug("Details to %s: %s" % (
                               self._taskName, traceback.format_exc()))


    @override
    def waitForCompletion(self, timeoutSec=None):
        doneFutures, _= futures.wait([self._future], timeout=timeoutSec)
        if self._future in doneFutures:
            self._printStackTraceInCaseOfError()
        else:
            raise TaskTimeoutError("Wait for completion"
                                   " of task '%s' failed" % self._taskName)


class Executor(object, metaclass=abc.ABCMeta):

    @abstractmethod
    def execute(self, task):
        assert False, "abstract base class"


class DoneTaskControl(TaskControl):

    @override
    def isDone(self):
        return True


    @override
    def isRunning(self):
        return False


    @override
    def waitForCompletion(self, timeoutSec):
            pass


class SerialExecutor(Executor):

    def __init__(self, logger=Logger.of("SerialExecutor")):
        Executor.__init__(self)
        self._logger= logger


    @override
    def execute(self, task):
        assert task is not None
        self._logger.debug("Executing task '%s'" % task.name())
        try:
            task.perform()
        except Exception as e:
            self._logger.error(str(e))

        return DoneTaskControl()


class MultiThreadExecutor(Executor):

    def __init__(self, maxThreads, logger=Logger.of("MultiThreadExecutor")):
        Executor.__init__(self)
        self._logger= logger
        self._executor= futures.ThreadPoolExecutor(max_workers=maxThreads)


    @override
    def execute(self, task):
        assert task is not None
        self._logger.debug("Executing task '%s'" % task.name())
        future= self._executor.submit(task.perform)
        return FutureTaskControl(future, task.name(), self._logger)


class SingleThreadExecutor(MultiThreadExecutor):

    def __init__(self, logger=Logger.of("SingleThreadExecutor")):
        MultiThreadExecutor.__init__(self, maxThreads=1, logger=logger)


class BusyExecutorException(Exception):

    def __init__(self, message, nameOfCurrentTask):
        Exception.__init__(self, message)
        self.nameOfCurrentTask= nameOfCurrentTask


class SingleThreadImmediateExecutor(Executor):

    def __init__(self, singleThreadExecutor):
        self._executor= singleThreadExecutor
        self._idle= True
        self._mutex= threading.RLock()
        self._currentTask= None


    @override
    @synchronized("_mutex")
    def execute(self, task):
        if not self._idle:
            taskName= self._currentTask.name()
            raise BusyExecutorException(
                "Executor is busy by task '%s'" % (taskName), taskName)

        class TrackedTask(Task):
            def __init__(self, task, onCompletionCallbackFunction):
                Task.__init__(self)
                self._task= task
                self._onCompletionCallback= onCompletionCallbackFunction

            @override
            def perform(self):
                try:
                    self._task.perform()
                finally:
                    self._onCompletionCallback()

            @override
            def name(self):
                return self._task.name()

        self._idle= False
        self._currentTask= task
        return self._executor.execute(TrackedTask(task, self._onCompletion))


    @synchronized("_mutex")
    def _onCompletion(self):
        self._currentTask= None
        self._idle= True


    @synchronized("_mutex")
    def isIdle(self):
        return self._idle


class DelayedTaskControl(TaskControl):


    def __init__(self):
        super(DelayedTaskControl, self).__init__()
        self._done= False


    @override
    def isDone(self):
        return self._done


    def markAsDone(self):
        self._done= True


    @override
    def isRunning(self):
        return False


    @override
    def waitForCompletion(self, timeoutSec):
        assert False, "N/A"


class TaskDelayingExecutor(Executor):


    def __init__(self):
        super(TaskDelayingExecutor, self).__init__()
        self._delayedTasks= []

    @override
    def execute(self, task):
        class WrappedTask(Task):
            def __init__(self, task, delayedTaskControl):
                super(WrappedTask, self).__init__()
                self._task= task
                self._taskControl= delayedTaskControl

            def perform(self):
                self._task.perform()
                self._taskControl.markAsDone()


        taskCtrl= DelayedTaskControl()
        self._delayedTasks.append(WrappedTask(task, taskCtrl))
        return taskCtrl


    def getDelayedTasks(self):
        return self._delayedTasks


    def executeDelayedTasks(self):
        for each in self._delayedTasks:
            each.perform()


class ExecutorFactory(object):


    @staticmethod
    def buildMultiThreadExecutor(maxNumThreads):
        return MultiThreadExecutor(maxNumThreads)


    @staticmethod
    def buildSingleThreadExecutor(logger=Logger.of("SingleThreadExecutor")):
        return SingleThreadExecutor(logger)


    @staticmethod
    def buildSingleThreadImmediateExecutor(
            logger=Logger.of("SingleThreadImmediateExecutor")):
        return SingleThreadImmediateExecutor(
            ExecutorFactory.buildSingleThreadExecutor(logger))



def logTaskFailure(func):
    @wraps(func)
    def wrappedMethod(self, *args, **kwds):
        try:
            return func(self, *args, **kwds)
        except Exception as e:
            message= "Task failed: %s" % str(e)
            self._logger.error(message)
            raise

    return wrappedMethod




# Python2/3 compatibility
from __future__ import print_function

# TODO add astropy units


def n_phot(
        mag,
        band=None,
        surf=1.0,
        delta_t=1.0,
        lambda_=None,
        width=None,
        e0=None,
        back_mag=None,
        model='default',
        verbose=False):
    '''
    This routine computes from a source magnitude the corresponding
    number of photons for a given band, a given surface and a given time
    interval. This is returned together with the number of photons from
    the sky background.
    The chosen band can be either a Johnson one (from U to M, but also
    a default Na-band), or a user-defined one (specifying the central
    wavelength and the bandwidth).
    The routine can also return the model used for calculations, if the
    attribute get_model() is called (see examples #4 and #5).

    TODO c in the code is approximated as 3e8

    INPUTS:

     mag  = magnitude [float]

    KEYWORDS:

     band     = wavelength Johnson band or Na default band [str].
     delta_t  = integrating time [s] [float], default is 1s.
     surf     = integrating surface [m^2] [float], default is 1m^2.
     lambda_   = central wavelenght of the choosen band [m] [float].
     width    = bandwidth of the choosen band [m] [float].
     e0       = A0 0-magnitude star brightness in the choosen band,
                [J/s/m^2/um] [float].
     back_mag = sky background default magnitude [FLOAT].
     model    = use the specified model instead the the default one (Lena 96).
                     Available model are: 'MAORY-1'   (phase A of Maory MCAO-EELT)

    OUTPUTS:

     Tuple with 2 elements:

     nb_of_photons = the computed source number of photons [float]
     background    = the computed sky background nb of photons per arcsec^2 [float]

    SIDE EFFECTS:
     none.

    RESTRICTIONS:
     none.

    EXAMPLES:
     [1]: compute the number of photons coming from a 5-mag star
          observed in V-band, with 8m-diameter telescope and an integration
          time of 12s:
 
          nb_of_photons, background = n_phot(5.,
                                             band='V',
                                             surf=math.pi*(8.**2)/4,
                                             delta_t=12.)
 
          returns:
             nb_of_photons = 5.9521160e+10 photons
             background    = 63581.948     photons
 
     [2]: compute the same stuff but with a user-defined sky background
          of 19.5-mag:
 
          nb_of_photons, background = n_phot(5.,
                                             band='V',
                                             surf=math.pi*(8.**2)/4,
                                             delta_t=12.,
                                             back_mag=19.5)
          returns:
             nb_of_photons = 5.9521160e+10 photons
             background    = 100549.18     photons
 
     [3]: compute the same stuff but with a user-defined band of
          central wavelength 0.54um and a narrow bandwidth of 0.01um:
 
          nb_of_photons, background = n_phot(5.,
                                             lambda_=5.4E-7,
                                             width=1E-8,
                                             surf=math.pi*(8.**2)/4,
                                             delta_t=12.,
                                             back_mag=19.5)
          returns:
             nb_of_photons = 6.5661736e+09 photons
             background    = 11092.246     photons
 
     [4]: get the sky background default magnitudes table:

          model = n_phot.getmodel()
          model.back_mag_table
 
 
          returns:
             back_mag_table= [22.,21.,20.,19.,17.5,16.,14.,12.,10.,6.,23.5]

 
     [5]: get the bands, central wavelengths, bandwidths, A0 0-mag. star
          brightnesses, and sky background default magnitudes tables:

          model = n_phot.getmodel()
 
          returns:
             model.band_table    =
 [   "U",   "B",   "V",   "R",   "I",   "J",   "H",   "K",   "L",   "M",    "Na"]
             model.lambda_table  =
 [3.6e-7,4.4e-7,5.5e-7,  7e-7,  9e-7,1.3e-6,1.7e-6,2.2e-6,3.4e-6,  5e-06,5.89e-7]
             model.width_table   =
 [6.8e-8,9.8e-8,8.9e-8,2.2e-7,2.4e-7,  3e-7,3.5e-7,  4e-7,5.5e-7,   3e-7,   1e-8]
             model.e0_table      =
 [4.4e-8,7.2e-8,3.9e-8,1.8e-8,8.3e-9,3.4e-9, 7e-10, 4e-10, 8e-11,2.2e-11, 3.9e-8]
             model.back_mag_table=
 [   23.,   22.,   21.,   20.,  19.5,   14.,  13.5,  12.5,    3.,     0.,    23.]
 
    RESTRICTIONS:
     in the case that user-defined bandwidth (keyword width) and
     wavelength (keyword lambda_) are set, it cannot extend beyond a
     routine-defined band. otherwise the results (nb_of_photons and
     background) are wrong.
 
    MODIFICATION HISTORY:
     program written: october 1998,
                        Marcel Carbillet (OAA) [marcel@arcetri.astro.it].
     modifications  : march 1999,
                        Marcel Carbillet (OAA) [marcel@arcetri.astro.it]:
                         - help completed.
                         - background output stuff added.
                      august 2015
                        Guido Agapito (OAA) [agapito@arcetri.astro.it]
                         - added DOUBLE keyword
                         - added verbose keyword
                         - where and closest like function to find the band/lambda
                      may 2019
                        Alfio Puglisi (OAA) [alfio.puglisi@inaf.it]
                         - translated to Python
                         - removed DOUBLE keyword
 

    '''
    mymodel = get_model(model)

    if band is None and lambda_ is None:
        raise ValueError('n_phot: cannot work if both band and lambda are None')

    # Get idx_band and lambda_ based on model tables,
    # and respect user overrides where possible.

    if band is not None:

        # Band given. Get lambda from table if not given.

        try:
            idx_band = mymodel.band_tab.index(band)
        except ValueError:
            raise Exception('n_phot: band %s not found in table' % band)

        if lambda_ is None:
            lambda_ = mymodel.lambda_tab[idx_band]

    else:

        # Band not given. Use lambda to index lambda table and find band

        if lambda_ < min(mymodel.lambda_tab):
            print('    ATTENTION:  lambda is < of the smallest element in lambda_tab')
        if lambda_ > max(mymodel.lambda_tab):
            print('    ATTENTION:  lambda is > of the greater element in lambda_tab')

        # Na band is chosen only if lambda is between 588.9 and 589.1 nm,
        # otherwise the band is searched in the other elements of lambda_tab
        # MAORY-1 model never uses Na.

        useNa = abs(lambda_ - 0.589e-6) < 0.0001e-6 and model != 'MAORY-1'
        if useNa:
            values = mymodel.lambda_tab
        else:
            values = mymodel.lambda_tab[0:-1]

        values = list(map(lambda x: abs(x - lambda_), values))
        idx_band = values.index(min(values))

    if verbose:
        print('BAND index number       : ', idx_band)
        print('BAND                    : ', mymodel.band_tab[idx_band])
        print('BAND CENTRAL WAVELENGTH : ', mymodel.lambda_tab[idx_band])

    if width is None:
        width = mymodel.width_tab[idx_band]  # bandwidth[m]

    if back_mag is None:
        back_mag = mymodel.back_mag_tab[idx_band]  # sky background magnitude

    if e0 is None:
        e0 = mymodel.e0_tab[idx_band]  # A0 star 0-mag. brightness [J/s/m^2/um]

    h = 6.626e-34  # Planck constant [Js]
    c = 3e8  # light velocity [m/s]

    # source number of photons
    nb_of_photons = lambda_ * delta_t * surf * (width * 1e6) * e0 / (h * c) * 10 ** (-mag / 2.5)

    # sky background nb of photons
    background = lambda_ * delta_t * surf * (width * 1e6) * e0 / (h * c) * 10 ** (-back_mag / 2.5)

    return (nb_of_photons, background)


class DefaultModel():
    '''
    Table of magnitudes for each band.

    ref.: P.Lena, Astrophysique : methodes physiques de l'observation,
            pp.95--96, Coll. Savoirs Actuels, InterEd./CNRS-Ed. (1996)

    except Na-band (added)
    '''

    # bands:
    band_tab = ['U', 'B', 'V', 'R', 'I', 'J', 'H', 'K', 'L', 'M', 'Na']

    # wavelength [um]:
    lambda_tab = \
      [0.36, 0.44, 0.55, 0.70, 0.90, 1.25, 1.65, 2.20, 3.40, 5.00, .589]

    # bandwidth [um]:
    width_tab = \
      [.068, .098, .089, .220, .240, .300, .350, .400, .550, .300, .010]

    # A0 0-magnitude star brightness [J/s/m^2/um] * 1e10
    e0_tab = \
      [435., 720., 392., 176., 83.0, 34.0, 7.00, 3.90, 0.81, 0.22, 392.]

    # default sky background magnitudes:
    back_mag_tab = \
      [23., 22., 21., 20., 19.5, 14., 13.5, 12.5, 3., 0., 23.]

    # list(map()) should work on both python2 and 3
    lambda_tab = list(map(lambda x: x * 1e-6, lambda_tab))
    width_tab = list(map(lambda x: x * 1e-6, width_tab))
    e0_tab = list(map(lambda x: x * 1e-10, e0_tab))


class Maory1Model():
    '''based on E-SPE-ESO-276-0206 issue 1 and MAORY Tech Notes 19 Dec 2007'''

    band_tab = ['V', 'R', 'I', ' J', 'H', 'K']
    lambda_tab = [0.54, 0.65, 0.80, 1.25, 1.65, 2.20]  # 1e-6
    width_tab = [.090, .130, .200, .300, .350, .400]  # 1e-6
    e0_tab = [368., 177., 119., 28.6, 10.8, 3.79]  # 1e-10
    back_mag_tab = [21.4, 20.6, 19.7, 16.5, 14.4, 13.0]

    # list(map()) should work on both python2 and 3
    lambda_tab = list(map(lambda x: x * 1e-6, lambda_tab))
    width_tab = list(map(lambda x: x * 1e-6, width_tab))
    e0_tab = list(map(lambda x: x * 1e-10, e0_tab))

# Add models as function attributes


n_phot.default_model = DefaultModel()
n_phot.maory_model = Maory1Model()


def get_model(model='default'):
    '''Callable function attribute to get the model's tables.'''
    if model == 'default':
        return DefaultModel()

    elif model == 'MAORY-1':
        return Maory1Model()

    else:
        raise Exception('n_phot: unknown model %s' % model)


n_phot.get_model = get_model

# Python2-compatible type hints

n_phot.__annotations__ = {
    'mag': float,
    'band': str,
    'surf': float,
    'delta_t': float,
    'lambda_': float,
    'width': float,
    'e0': float,
    'back_mag': float,
    'model': str,
    'verbose': bool,
}







class PickelsLibrary(object):

    @classmethod
    def baseUrl(cls):
        return 'http://ssb.stsci.edu/cdbs/grid/pickles/dat_uvk/'


    @classmethod
    def filename(cls, spectralType):
        """
        Return URL to access Pickels UVKLIB spectra at STScI
        To be used with synphot.SourceSpectrum.from_file()
        """
        return "%s%s.fits" % (cls.baseUrl(),
                              cls._spectralTypeDict()[spectralType])

    @classmethod
    def _spectralTypeDict(cls):
        return {
            "O5V": "pickles_uk_1",
            "O9V": "pickles_uk_2",
            "B0V": "pickles_uk_3",
            "B1V": "pickles_uk_4",
            "B3V": "pickles_uk_5",
            "B5-7V": "pickles_uk_6",
            "B8V": "pickles_uk_7",
            "A0V": "pickles_uk_9",
            "A2V": "pickles_uk_10",
            "A3V": "pickles_uk_11",
            "A5V": "pickles_uk_12",
            "F0V": "pickles_uk_14",
            "F2V": "pickles_uk_15",
            "F5V": "pickles_uk_16",
            "F8V": "pickles_uk_20",
            "G0V": "pickles_uk_23",
            "G2V": "pickles_uk_26",
            "G5V": "pickles_uk_27",
            "G8V": "pickles_uk_30",
            "K0V": "pickles_uk_31",
            "K2V": "pickles_uk_33",
            "K3V": "pickles_uk_34",
            "K5V": "pickles_uk_36",
            "K7V": "pickles_uk_37",
            "M0V": "pickles_uk_38",
            "M2V": "pickles_uk_40",
            "M4V": "pickles_uk_43",
            "M5V": "pickles_uk_44",
            "B2IV": "pickles_uk_46",
            "B6IV": "pickles_uk_47",
            "A0IV": "pickles_uk_48",
            "A4-7IV": "pickles_uk_49",
            "F0-2IV": "pickles_uk_50",
            "F5IV": "pickles_uk_51",
            "F8IV": "pickles_uk_52",
            "G0IV": "pickles_uk_53",
            "G2IV": "pickles_uk_54",
            "G5IV": "pickles_uk_55",
            "G8IV": "pickles_uk_56",
            "K0IV": "pickles_uk_57",
            "K1IV": "pickles_uk_58",
            "K3IV": "pickles_uk_59",
            "O8III": "pickles_uk_60",
            "B1-2III": "pickles_uk_61",
            "B5III": "pickles_uk_63",
            "B9III": "pickles_uk_64",
            "A0III": "pickles_uk_65",
            "A5III": "pickles_uk_67",
            "F0III": "pickles_uk_69",
            "F5III": "pickles_uk_71",
            "G0III": "pickles_uk_72",
            "G5III": "pickles_uk_73",
            "G8III": "pickles_uk_76",
            "K0III": "pickles_uk_78",
            "K3III": "pickles_uk_87",
            "K5III": "pickles_uk_93",
            "M0III": "pickles_uk_95",
            "M5III": "pickles_uk_100",
            "M10III": "pickles_uk_105",
            "B2II": "pickles_uk_106",
            "B5II": "pickles_uk_107",
            "F0II": "pickles_uk_108",
            "F2II": "pickles_uk_109",
            "G5II": "pickles_uk_110",
            "K0-1II": "pickles_uk_111",
            "K3-4II": "pickles_uk_112",
            "M3II": "pickles_uk_113",
            "B0I": "pickles_uk_114",
            "B5I": "pickles_uk_117",
            "B8I": "pickles_uk_118",
            "A0I": "pickles_uk_119",
            "F0I": "pickles_uk_121",
            "F5I": "pickles_uk_122",
            "F8I": "pickles_uk_123",
            "G0I": "pickles_uk_124",
            "G5I": "pickles_uk_126",
            "G8I": "pickles_uk_127",
            "K2I": "pickles_uk_128",
            "K4I": "pickles_uk_130",
            "M2I": "pickles_uk_131",
        }


for k in PickelsLibrary._spectralTypeDict().keys():
    globals()[k]= PickelsLibrary.filename(k)

